#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Tests for ProductionData."""
from __future__ import absolute_import
import unittest

from mock import patch, MagicMock as Mock
from DIRAC import S_OK, S_ERROR
from ILCDIRAC.Core.Utilities.ProductionData import constructUserLFNs, getExperimentFromPath
from Tests.Utilities.GeneralUtils import assertDiracSucceedsWith, assertDiracSucceedsWith_equals

__RCSID__ = "$Id$"

MODULE_NAME = 'ILCDIRAC.Core.Utilities.ProductionData'


class TestProductionData(unittest.TestCase):
  """Tests constructUserLFNs."""

  # Rest of class is covered by other tests

  def test_construct_user_lfns(self):
    today_mock = Mock()
    today_mock.timetuple.return_value = [1999, 27]
    date_mock = Mock()
    date_mock.today.return_value = today_mock
    with patch('%s.datetime.date' % MODULE_NAME, new=date_mock), \
         patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value='testpre')):
      result = constructUserLFNs(1234567, 'mytestVO', 'mytestowner',
                                  ['myTestOutputFile1.txt', 'myTestOutputFile2.ppt'],
                                  '/mydir/dir//MyTestOutputPath1')
      assertDiracSucceedsWith(
          result, '/mytestVO/testpre/m/mytestowner/mydir/dir/MyTestOutputPath1/myTestOutputFile1.txt', self)
      assertDiracSucceedsWith(
          result, '/mytestVO/testpre/m/mytestowner/mydir/dir/MyTestOutputPath1/myTestOutputFile2.ppt', self)

  def test_construct_user_lfns_getvo_fails(self):
    today_mock = Mock()
    today_mock.timetuple.return_value = [1999, 27]
    date_mock = Mock()
    date_mock.today.return_value = today_mock
    with patch('%s.datetime.date' % MODULE_NAME, new=date_mock), \
         patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value='testpre')), \
         patch('%s.getVOfromProxyGroup' % MODULE_NAME, new=Mock(return_value=S_ERROR('some_err'))), \
         patch('%s.gLogger.error' % MODULE_NAME) as log_mock:
      result = constructUserLFNs(1234567, None, 'mytestowner',
                                  ['myTestOutputFile1.txt', 'myTestOutputFile2.ppt'],
                                  '/mydir/dir//MyTestOutputPath1')
      assertDiracSucceedsWith(
          result, '/ilc/testpre/m/mytestowner/mydir/dir/MyTestOutputPath1/myTestOutputFile1.txt', self)
      assertDiracSucceedsWith(
          result, '/ilc/testpre/m/mytestowner/mydir/dir/MyTestOutputPath1/myTestOutputFile2.ppt', self)
      log_mock.assert_called_once_with('Could not get VO from CS, assuming ilc')

  def test_construct_user_lfns_novo(self):
    today_mock = Mock()
    today_mock.timetuple.return_value = [1999, 27]
    date_mock = Mock()
    date_mock.today.return_value = today_mock
    with patch('%s.datetime.date' % MODULE_NAME, new=date_mock), \
         patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value='testpre')), \
         patch('%s.getVOfromProxyGroup' % MODULE_NAME, new=Mock(return_value=S_OK('mytestproxyvo'))):
      result = constructUserLFNs(1234567, None, 'mytestowner',
                                  ['myTestOutputFile1.txt', 'myTestOutputFile2.ppt'],
                                  '/mydir/dir//MyTestOutputPath1')
      assertDiracSucceedsWith(
          result, '/mytestproxyvo/testpre/m/mytestowner/mydir/dir/MyTestOutputPath1/myTestOutputFile1.txt', self)
      assertDiracSucceedsWith(
          result, '/mytestproxyvo/testpre/m/mytestowner/mydir/dir/MyTestOutputPath1/myTestOutputFile2.ppt', self)

  def test_construct_user_lfns_no_outputpath(self):
    today_mock = Mock()
    today_mock.timetuple.return_value = [1999, 27]
    date_mock = Mock()
    date_mock.today.return_value = today_mock
    with patch('%s.datetime.date' % MODULE_NAME, new=date_mock), \
         patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value='testpre')):
      result = constructUserLFNs(1234567, 'mytestVO', 'mytestowner', '/ignore/this/myOutputString.md5', None)
      assertDiracSucceedsWith(
          result, '/mytestVO/testpre/m/mytestowner/1999_27/1234/1234567/myOutputString.md5', self)

  def test_construct_user_lfns_no_lfns(self):
    with patch('%s.Operations.getValue' % MODULE_NAME, new=Mock(return_value='testpre')), \
        patch('%s.gLogger.info' % MODULE_NAME) as log_mock:
      result = constructUserLFNs(1234567, 'mytestVO', 'mytestowner', [], '')
      log_mock.assert_called_once_with('No output LFN(s) constructed')
      assertDiracSucceedsWith_equals(result, [], self)

  def test_getExperimentFromBasePath(self):
    optDict = S_OK({'exp1': '/ilc/prod/exp1', 'exp2': '/ilc/prod/exp2/path1, /ilc/prod/exp2/path2'})
    with patch('%s.Operations.getOptionsDict' % MODULE_NAME, new=Mock(return_value=optDict)):
      experiment = getExperimentFromPath(Mock(), '/ilc/prod/exp2/path1/log/file', 'default')
      self.assertEqual(experiment, 'exp2')
      experiment = getExperimentFromPath(Mock(), '/ilc/prod/exp/log/file', 'default')
      self.assertEqual(experiment, 'default')

    optDict = S_ERROR("not found")
    with patch('%s.Operations.getOptionsDict' % MODULE_NAME, new=Mock(return_value=optDict)):
      experiment = getExperimentFromPath(Mock(), '/ilc/prod/exp2/path1/log/file', 'default')
      self.assertEqual(experiment, 'default')



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""The job path resolution module is a VO-specific plugin that allows to define VO job policy in a simple way.  This allows the inclusion of ILC specific WMS optimizers without compromising the generic nature of DIRAC.

The arguments dictionary from the JobPathAgent includes the ClassAd
job description and therefore decisions are made based on the existence
of JDL parameters.
:author: Stuart Paterson
"""

from __future__ import absolute_import
__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gLogger

LOG = gLogger.getSubLogger(__name__)
COMPONENT_NAME = 'ILCJobPathResolution'


class JobPathResolution(object):
  """VO Specific plugin for jobpath resolution."""
  #############################################################################

  def __init__(self, argumentsDict):
    """Standard constructor."""
    self.arguments = argumentsDict
    self.name = COMPONENT_NAME

  #############################################################################
  def execute(self):
    """Given the arguments from the JobPathAgent, this function resolves job optimizer paths according to ILC VO policy."""
    ilcPath = ''

    if 'ConfigPath' not in self.arguments:
      LOG.warn('No CS ConfigPath defined')
      return S_ERROR('JobPathResoulution Failure')

    LOG.verbose('Attempting to resolve job path for ILC')
    job = self.arguments['JobID']
    classadJob = self.arguments['ClassAd']

    inputData = classadJob.get_expression('InputData').replace('"', '').replace('Unknown', '')
    if inputData:
      LOG.info('Job %s has input data requirement' % (job))
      ilcPath += 'InputData'

    if not ilcPath:
      LOG.info('No ILC specific optimizers to be added')

    return S_OK(ilcPath)

#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""FCC application is defined by 2 modules which are :

- ILCDIRAC.Interfaces.API.NewInterface.Applications.Fcc (this module)
- ILCDIRAC.Workflow.Modules.FccAnalysis

FCC applications can run under a UserJob located here:

- ILCDIRAC.Interfaces.API.NewInterface.UserJob
"""

# standard libraries
from __future__ import absolute_import
import os
import re
import shutil
import types

# DIRAC libraries
from ILCDIRAC.Interfaces.API.NewInterface.Application import Application
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Workflow.Parameter import Parameter

LOG = gLogger.getSubLogger(__name__)
__RCSID__ = "$Id$"


class Fcc(Application):
  """Fcc class defines the skeleton of an FCC application.

  It inherits from the Application module.

  Fcc application is the base class for :
  -   FccSw
  -   FccAnalysis

  Look their definition at the end of this module.
  """

  def __init__(self):

    # Required
    self.fccExecutable = ''

    # Path of the FCCSW installation (Local or in CVMFS)
    self.fccSwPath = ''

    # FccSw generates an additionnal configuration file
    self.isGaudiOptionsFileNeeded = False

    # Read or generate events
    self.read = False

    # Random generator service used to set the seed and the number of events
    self.randomGenerator = {}

    # Final FCC input sandbox
    self._inputSandbox = set()

    # Temporary FCC input sandbox
    # contains user files/folders not yet checked
    self._tempInputSandbox = set()

    # Some folders have to be filtered (like 'Detector' folder of FCCSW installation)
    # to avoid sandbox overload (sandbox max size = 10 Mb)
    self._foldersToFilter = []

    # Folder filters
    # which extension to filter
    self._filteredExtensions = []

    # What operation to do for each extension (include it or exclude it)
    self._excludesOrIncludes = []

    # Local path of the temporary sandbox
    self._tempCwd = os.path.realpath('temp_fcc_dirac')

    # Output sandbox (not the real one)
    # used just for printing informations to the user
    self._outputSandbox = set()

    # Default application log level
    self.logLevel = "DEBUG"

    self.datatype = 'REC'
    self.detectortype = 'ILD'

    super(Fcc, self).__init__()
    # Those 6 need to come after default constructor
    self._modulename = 'FccAnalysis'
    #self._importLocation = 'ILCDIRAC.Workflow.Modules'
    self._moduledescription = "Module running FCC software"
    self.appname = self.__class__.__name__
    self.version = "v0.8.1"
    self.energy = 0
    self.numberOfEvents = 0

  def _applicationModule(self):
    """It transfers parameter names from the module Fcc to the module FccAnalysis.

    :return: The module that takes these parameters
    :rtype: moduleinstance
    """

    md1 = self._createModuleDefinition()

    md1.addParameter(Parameter("fccExecutable", "", "string", "", "", False, False,
                               "The executable to run"))

    md1.addParameter(Parameter("isGaudiOptionsFileNeeded", False, "bool", "", "", False, False,
                               "Gaudi configuration file"))

    md1.addParameter(Parameter("logLevel", "", "string", "", "", False, False,
                               "Gaudi Log Level"))

    md1.addParameter(Parameter("read", "", "string", "", "", False, False,
                               "Application can read or generate events"))

    md1.addParameter(Parameter("randomGenerator", {}, "dict", "", "", False, False,
                               "Pythia card files"))

    return md1

  def _applicationModuleValues(self, moduleinstance):
    """It transfers parameter values from the module Fcc to the module FccAnalysis.

    :param moduleinstance: The module we load (FccAnalysis)
    """

    moduleinstance.setValue("fccExecutable", self.fccExecutable)
    moduleinstance.setValue("isGaudiOptionsFileNeeded", self.isGaudiOptionsFileNeeded)
    moduleinstance.setValue("logLevel", self.logLevel)
    moduleinstance.setValue("read", self.read)
    moduleinstance.setValue("randomGenerator", self.randomGenerator)

  def _checkConsistency(self, job=None):
    """This function checks the minimum requirements of the application and updates the sandbox with files/folders required by the application.

    :param job: The job containing the application
    :type job: DIRAC.Interfaces.API.Job.Job

    :return: The success or failure of the consistency checking
    :rtype: DIRAC.S_OK, DIRAC.S_ERROR
    """

    infoMessage = (
        "Application general consistency :"
        " _checkConsistency() on '%(name)s'..." % {'name': self.appname}
        )
    LOG.info(infoMessage)

    if not self.version:
      errorMessage = 'Version not set!'
      LOG.error(errorMessage)
      return S_ERROR(errorMessage)

    infoMessage = (
        "Consistency : Version of the application set to :"
        "\nversion : %(version)s" % {'version': self.version}
        )
    LOG.info(infoMessage)

    # The executable is mandatory and also the configuration file except if the
    # input is taken from an other app hence the use of '_inputapp'
    if not (self.fccExecutable and (self.steeringFile or self._inputapp)):
      errorMessage = (
          "Consistency : Error in parsing '%(name)s' application :\n"
          "You have to provide at least an executable"
          " and a configuration file for each application" % {'name': self.appname}
          )
      LOG.error(errorMessage)
      return S_ERROR(errorMessage)

    if not isinstance(self.steeringFile, str):
      errorMessage = (
          "Consistency : Fcc Application accepts only one input configuration file:\n"
          "If you want to run the application '%(name)s' with many configurations then\n"
          "Create an new application with the other configuration\n"
          "You can also use the 'getInputFromApp' method to link applications" % {'name': self.appname}
          )
      LOG.error(errorMessage)
      return S_ERROR(errorMessage)

    debugMessage = (
        "Consistency : Executable and configuration of the application set to :"
        "\nexecutable : %(exec)s"
        "\nconfiguration : %(conf)s" % {'exec': self.fccExecutable, 'conf': self.steeringFile}
        )
    LOG.debug(debugMessage)

    # All input files are put in the FCC temporary sandbox for a
    # pre-checking before being added to the FCC final sandbox
    if self.steeringFile:
      self._tempInputSandbox.add(self.steeringFile)
      # Once steeringFile is given, we set the random generator for 'fcc-physics-read'
      if not self.read and "Pythia" in self.randomGenerator:
        self.randomGenerator["Pythia"].append(os.path.basename(self.steeringFile))

    LOG.info("Sandboxing : Sandboxing in progress...")

    # We update the sandbox with files/folders required by the application
    if not self._importToSandbox():
      errorMessage = "_importToSandbox() failed"
      LOG.error(errorMessage)
      return S_ERROR(errorMessage)

    LOG.info("Sandboxing : Sandboxing successfull")

    # setOutputFile() method informs the job that this application has an output file
    # This output can be used as input for another application.
    # In this way, app2.getInputFromApp(app1) method knows the ouput file of the given application
    # app1 thanks to its method setOutputFile().

    # self.inputSB is an attribute of the DIRAC Application and not of FCC.
    # The description file of the job (JDL file) contains a section for the input sandbox
    # This section is filled with a list of files (self.inputSB).
    # After user input files, application files, and application additionnal
    # files checked in the temporary sandbox, we set
    # the DIRAC application input sandbox : self.inputSB

    self._inputSandbox = self._inputSandbox.union(self._foldersToFilter)
    self.inputSB = list(self._inputSandbox)

    # Sandbox can be set at the application level or at the job level.
    # Whatever the level choosed, sandbox files are all put
    # in the same final destination which is a list of paths
    # in the JDL file (see Input Sandbox parameter of the JDL file).

    infoMessage = (
        "Application general consistency : _checkConsistency()"
        " on '%(name)s' successfull" % {'name': self.appname}
        )
    LOG.info(infoMessage)

    return S_OK(infoMessage)

  def _checkFinalConsistency(self):
    """

    :return: The success of the final consistency checking
    :rtype: DIRAC.S_OK

    """

    # Many output files can be managed if setOutputFile() method accepts list
    if self.outputFile:
      outputFiles = [self.outputFile] if isinstance(self.outputFile, str) else self.outputFile
      for outputFile in outputFiles:
        self._outputSandbox.add(
            "%s(_JobID).root (%s)" %
            (os.path.splitext(outputFile)[0],
             "Name of the eventual output root file"))

    # We add the log file and the output file to the output sandbox
    self._outputSandbox.add(self.logFile)

    infoMessage = (
        "\n********************************FCC SUMMARY******************************\n"
        "You plan to submit this application with its corresponding log :\n"
        "%(name)s --> %(log)s" % {'name': self.appname, 'log': self.logFile}
        )
    LOG.info(infoMessage)

    if self._inputSandbox:
      infoMessage = (
          "\nHere is the content of its input sandbox :\n%(input)s"
          % {'input': '\n'.join(self._inputSandbox)}
          )
      LOG.info(infoMessage)

    if self._outputSandbox:
      infoMessage = (
          "\nHere are the output files :\n%(output)s"
          % {'output': '\n'.join(self._outputSandbox)}
          )
      LOG.info(infoMessage)

    infoMessage = "\n********************************FCC SUMMARY******************************"
    LOG.info(infoMessage)

    return super(Fcc, self)._checkFinalConsistency()

  def _checkWorkflowConsistency(self):
    """Summary of the application done in _checkConsistency() method.

    :return: The success or the failure of _checkRequiredApp()
    :rtype: DIRAC.S_OK, DIRAC.S_ERROR
    """
    return self._checkRequiredApp()

  def _prodjobmodules(self, stepdefinition):

    # Here one needs to take care of listoutput
    if self.outputPath:
      self._listofoutput.append({'OutputFile': '@{OutputFile}', "outputPath": "@{OutputPath}",
                                 "outputDataSE": '@{OutputSE}'})

    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setOutputComputeDataList(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('prodjobmodules failed')
    return S_OK()

  def _resolveLinkedStepParameters(self, stepinstance):
    if isinstance(self._linkedidx, int):
      self._inputappstep = self._jobsteps[self._linkedidx]
    if self._inputappstep:
      stepinstance.setLink("InputFile", self._inputappstep.getType(), "OutputFile")
    return S_OK()

  def _userjobmodules(self, stepdefinition):
    res1 = self._setApplicationModuleAndParameters(stepdefinition)
    res2 = self._setUserJobFinalization(stepdefinition)
    if not res1["OK"] or not res2["OK"]:
      return S_ERROR('userjobmodules failed')
    return S_OK()

###############################  Fcc METHODS #####################################################
  @staticmethod
  def _findPath(path):
    """This function checks if file/folder exists.

    :param path: The path to look for
    :type path: str

    :return: The absolute path and its existence
    :rtype: (str, bool)
    """

    tempPath = os.path.realpath(path)
    return (tempPath, True) if os.path.exists(tempPath) else (path, False)

  def _importFiles(self):
    """This function adds folders/files to the sandbox specified by the user.

    :return: The success or the failure of the import
    :rtype: bool
    """

    if not self._tempInputSandbox:
      warnMessage = "Sandboxing : Your application has an empty input sandbox"
      LOG.warn(warnMessage)
      return True

    for path in self._tempInputSandbox:
      # We make a pre-checking of files in reachable filesystems (e.g. AFS, CVMFS)
      path, exists = self._findPath(path)

      # If file does not exist then consistency fails
      if not exists:
        errorMessage = (
            "Sandboxing : The path '%(path)s' does not exist\n"
            "Please ensure that your path exists in an accessible file system "
            "(AFS or CVMFS)" % {'path': path}
            )
        LOG.error(errorMessage)
        return False

      if path.startswith('/afs/'):
        warnMessage = (
            "Sandboxing : You plan to upload '%(path)s'"
            " which is stored on AFS\n"
            "STORING FILES ON AFS IS DEPRECATED" % {'path': path}
            )

        # We log the message into the warning level
        LOG.warn(warnMessage)

      # cvmfs paths do not need to be uploaded, they can be accessed remotely.
      # but for the moment do not be smart about it, add also cvmfs files,
      # no need to check.
      # if not path.startswith('/cvmfs/'):
      # if path is already in the sandbox, set type will kill duplicates
      self._inputSandbox.add(path)

      debugMessage = (
          "Sandboxing : The path '%(path)s' required by the application"
          " has been added to te sandbox" % {'path': path}
          )
      LOG.debug(debugMessage)

    debugMessage = (
        "Sandboxing : Files required by FCC application"
        " verified and added successfully to the sandbox"
        )

    LOG.debug(debugMessage)
    return True

  def _importToSandbox(self):
    """This function checks all the files and folders of the FCC temporary sandbox and add them to the FCC 'final' sandbox.

    :return: The success or the failure of the import
    :rtype: bool
    """

    LOG.debug("Sandboxing : Importation of user files/folders...")

    # Import files required by the application
    # If import process fails for some reasons (see functions above for more details)
    # then consistency fails
    if not self._importFiles():
      LOG.error("Sandboxing : _importFiles() failed")
      return False
      # Do not continue remaining checks

    LOG.debug("Sandboxing : Importation of user files/folders successfull")

    return True

  @staticmethod
  def _readFromFile(fileName):
    """This function reads a file and returns its content.

    :param fileName: The path of the file to read
    :type fileName: str

    :return: The content of the file
    :rtype: str, str
    """

    try:
      with open(fileName, 'r') as fileToRead:
        content = fileToRead.read()
    except IOError as e:
      return None, 'Sandboxing : FCC file reading failed\n%s' % e

    return content, 'Sandboxing : FCC file reading successfull'

###############################  Fcc DAUGHTER CLASSES ##############################################


class FccSw(Fcc):
  """Definition of an FCCSW application.

  Usage:

  >>> fccSw = FccSw(
      fccConfFile='/build/username/FCC/FCCSW/Examples/options/geant_pgun_fullsim.py',
      fccSwPath='/build/username/FCC/FCCSW'
    )
  """

  def __init__(self, fccConfFile="", fccSwPath="", read=False):

    super(FccSw, self).__init__()

    self.fccSwPath = os.path.realpath(fccSwPath)
    self.steeringFile = fccConfFile
    self.read = read

  def _checkConsistency(self, job=None):
    """Redefinition of ILCDIRAC.Workflow.Modules.Fcc._checkConsistency().

    :param job: The job containing the application
    :type job: DIRAC.Interfaces.API.Job.Job

    :return: The success or the failure of the consistency checking
    :rtype: DIRAC.S_OK, DIRAC.S_ERROR
    """

    LOG.debug("FCCSW specific consistency : _checkConsistency()...")

    if not self.fccSwPath or not os.path.exists(self.fccSwPath):
      errorMessage = (
          "FCCSW specific consistency : Error in parsing FCCSW application :\n"
          "You have to provide a valid path of the FCCSW installation"
          )
      LOG.error(errorMessage)
      return S_ERROR(errorMessage)

    debugMessage = (
        "FCCSW specific consistency : Creation of a temporary folder 'temp_fcc_dirac'"
        " in the current working directory..."
        )
    LOG.debug(debugMessage)

    # First, it creates a temporary local directory for folders whose
    # the content does not have to be sandboxed entirely
    # like filtered folders
    if not os.path.exists(self._tempCwd):
      try:
        os.makedirs(self._tempCwd)
      except OSError as e:
        errorMessage = "FCCSW specific consistency : Creation of 'temp_fcc_dirac' folder failed\n%s" % e
        LOG.error(errorMessage)
        return S_ERROR(errorMessage)

      LOG.debug("FCCSW specific consistency : Creation of 'temp_fcc_dirac' folder successfull")
    else:
      LOG.debug("FCCSW specific consistency : The temporary folder 'temp_fcc_dirac' already exists")

    # InstallArea folder is present on CVMFS so nothing to do
    # else if local FCCSW installation is used
    # download it because 'FCCSW.xenv' needs libraries from this folder

    self.fccExecutable = '%s/run gaudirun.py' % self.fccSwPath

    if not self.fccSwPath.startswith('/cvmfs/'):
      # 'FCCSW/run' script calls another 'run' script inside 'build.$BINARY_TAG' folder
      # The command works with a '.xenv' file located into 'build.$BINARY_TAG' relative to files of the cwd
      # We read the command of the 'run' script and we use the 'xenv' file of InstallArea instead
      # which is only related to CVMFS
      # If the reading of the command failed, we execute by default the command of the 0.8.1 release

      #python = '$PYTHON_BIN'
      python = '/cvmfs/sft.cern.ch/lcg/views/LCG_88/x86_64-slc6-gcc62-opt/bin/python'
      #xenv = '$XENV'
      xenv = '/cvmfs/fcc.cern.ch/sw/0.8.1/gaudi/v28r2/x86_64-slc6-gcc62-opt/scripts/xenv'
      argXenv = 'InstallArea/FCCSW.xenv'

      executablePath = ''

      for path in os.listdir(self.fccSwPath):
        absPath = os.path.join(self.fccSwPath, path)
        if path.startswith('build.') and not os.path.isfile(absPath):
          executablePath = os.path.join(absPath, 'run')

      if executablePath and os.path.exists(executablePath):
        executableContent, message = self._readFromFile(executablePath)

        if not executableContent:
          LOG.warn(message)
          LOG.debug('FCCSW specific consistency : Using by default the command of the 0.8.1 release !')
        else:
          LOG.debug(message)

          # Separate shebang (0) and command of the executable (1)
          executableCommand = executableContent.split('\n')[1]
          # Break command of the executable into parts
          executableParts = executableCommand.split()

          python = executableParts[1]
          xenv = executableParts[2]

      self.fccExecutable = 'exec %s %s --xml %s gaudirun.py' % (python, xenv, argXenv)
      installAreaFolder = os.path.join(self.fccSwPath, 'InstallArea')
      # We do not need all the content of these folders hence the filtering
      self._foldersToFilter.append(installAreaFolder)

      # Explanation
      # InstallAreaFolder : all dbg files are excluded
      # detectorFolder : only xml files are included
      self._filteredExtensions += ['.dbg']
      self._excludesOrIncludes += [True]

    # 'Detector' folder is not present on CVMFS in release v0.8.1 so the user has to add this
    # folder to the sandbox ('cf the corresponding warning message')
    # Actually, FCCSW installation of CVMFS run successfully only with configuration files that
    # do not need additionnal files like 'Generation/data/ParticleTable.txt' or folders like 'Detector'
    # FCCSW release made after 31/08/2017 will put a complete installation of FCCSW
    # And all examples of Examples/options should run successfully

    LOG.debug("FCCSW specific consistency : _checkConsistency() successfull")

    return super(FccSw, self)._checkConsistency()

  def _checkFinalConsistency(self):
    """Redefinition of ILCDIRAC.Workflow.Modules.Fcc._checkFinalConsistency(). Setting True 'isGaudiOptionsFileNeeded' attribute tells that the application needs to generate the gaudi option file of FccSw.

    :return: The success of the final consistency checking
    :rtype: DIRAC.S_OK
    """

    self.isGaudiOptionsFileNeeded = True
    return super(FccSw, self)._checkFinalConsistency()

  def _importToSandbox(self):
    """Redefinition of FCC._importToSandbox() method. FCCSW needs extra folders like 'InstallArea', 'Detector' and extra files specified in its configuration file.

    :return: The success or the failure of the import
    :rtype: bool
    """

    if not super(FccSw, self)._importToSandbox():
      errorMessage = "Sandboxing : _importToSandbox() failed"
      LOG.error(errorMessage)
      return False

    if not self._importFccswFiles():
      return False

    # After sandboxing, we filter some folders required by applications
    # and we import the filtered folders to the sandbox.
    if not self._setFilterToFolders():
      errorMessage = "_setFilterToFolders() failed"
      LOG.error(errorMessage)
      return False

    return True

  def _importFccswFiles(self):
    """FCCSW application needs additional files specified in the configuration file It also needs folders like 'InstallArea' and 'Detector'.

    :return: The success or the failure of the import
    :rtype: bool
    """

    # 'InstallArea' Folder already resolved and added in FccSw class
    # It is present in CVMFS contrary to 'Detector' folder
    #installAreaFolder = os.path.join(self.fccSwPath, 'InstallArea')
    detectorFolder = os.path.join(self.fccSwPath, 'Detector')

    # As CVMFS installation is not complete ('Detector' folder missing in release v0.8.1)
    # we do this check
    if not os.path.exists(detectorFolder):
      warnMessage = (
          "Sandboxing : The folder 'Detector' does not exist,"
          " it is not present in the FCCSW installation"
          "\nThen you should have added it manually to the input sandbox !"
          )
      LOG.warn(warnMessage)
    else:
      # We do not need all the content of these folders hence the filtering
      self._foldersToFilter.append(detectorFolder)

      # Explanation
      # InstallAreaFolder : all dbg files are excluded
      # detectorFolder : only xml files are included
      self._filteredExtensions += ['.xml']
      self._excludesOrIncludes += [False]

    LOG.debug("Sandboxing : FCC configuration file reading...")

    content, message = self._readFromFile(self.steeringFile)

    # If the configuration file is not valid then consistency fails
    if not content:
      LOG.error(message)
      return False

    LOG.debug(message)

    # Find all additional files specified in the fccsw configuration file
    #xml_files = re.findall(r'file:(.*.xml)',content)

    txtFiles = re.findall(r'(.*) *= *"(.*.txt)"', content)
    cmdFiles = re.findall(r'(.*) *= *"(.*.cmd)"', content)

    # Upload file not commented
    txtFiles = [txtFile[1] for txtFile in txtFiles if not txtFile[0].startswith("#")]
    cmdFiles = [cmdFile[1] for cmdFile in cmdFiles if not cmdFile[0].startswith("#")]

    lookForPythia = re.findall(r'.* *= *PythiaInterface *\(', content)

    # Check if PythiaInterface is instantiated somewhere and not commented
    isPythiaGeneratorUsed = True if lookForPythia and not lookForPythia[0].startswith("#") else False

    if isPythiaGeneratorUsed and cmdFiles:
      self.randomGenerator["Pythia"] = cmdFiles
    else:
      self.randomGenerator["Gaudi"] = True

    # From these paths we re-create the tree in the temporary sandbox
    # with only the desired file.
    # In the configuration file, these paths are relative to FCCSW installation.
    # e.g. Generation/data/foo.xml

    if not self._resolveTreeOfFiles(txtFiles, '.txt'):
      LOG.error("Sandboxing : _resolveTreeOfFiles() failed")
      return False
      # Do not continue remaining checks

    # We do the same now for '.cmd' files that may be specified in the configuration file
    return self._resolveTreeOfFiles(cmdFiles, '.cmd')

  def _resolveTreeOfFiles(self, files, extension):
    """Resolve FCC Configuration Files.

    FCC configuration file like 'Examples/options/geant_pgun_fullsim.py' needs files coming from FCCSW installation.
    The path of these files are hard-coded in the FCC configuration file with a relative path to FCCSW installation.

    This function aims to resolve the absolute path of each hard-coded files
    required by the configuration file.

    Once we have the absolute path of the file, we recreate the tree strucutre
    of the file in a temporary folder and copy the file to it.

    Because the original directory of the file may contain other files, in this
    way we copy only the desired file.

    The file has now a new location and this is this new tree that will be
    added to the sandbox.

    :param files: The files specified in the configuration file
    :type files: list

    :param extension: The extension of file to resolve
    :type extension: str

    :return: success or failure of the file copy
    :rtype: bool
    """

    if not files:
      warnMessage = (
          "Sandboxing : FCCSW configuration file"
          " does not seem to need any additional '%(ext)s' files" % {'ext': extension}
          )
      LOG.warn(warnMessage)
      return True

    for afile in files:

      source = os.path.realpath(os.path.join(self.fccSwPath, afile))
      destination = os.path.realpath(os.path.join(self._tempCwd, afile))

      if not os.path.exists(source):
        warnMessage = (
            "Sandboxing : The file '%s' does not exist,"
            " it is not present in the FCCSW installation"
            "\nThen you should have added it manually to the input sandbox !" % {'source': source}
            )
        LOG.warn(warnMessage)
        continue

      # We save the relative path of the file
      # e.g. Generation/data/
      tree = os.path.dirname(afile)
      # We prepend the temporary sandbox to the tree
      # which will become the new location of the file.
      treeFullPath = os.path.join(self._tempCwd, tree)

      debugMessage = (
          "Sandboxing : Tree '%(tree)s' of additionnal"
          " '%(ext)s' files creation..." % {'tree': treeFullPath, 'ext': extension}
          )
      LOG.debug(debugMessage)

      # We create the tree locally in the temporary folder
      if not os.path.exists(treeFullPath):
        try:
          os.makedirs(treeFullPath)
        except OSError as e:
          errorMessage = (
              "Sandboxing : Tree '%(tree)s' of additionnal"
              " '%(ext)s' files creation failed\n%(error)s" % {'tree': treeFullPath, 'ext': extension, 'error': e}
              )
          LOG.error(errorMessage)
          return False

        debugMessage = (
            "Sandboxing : Tree '%(tree)s' of additionnal"
            " '%(ext)s' files creation successfull" % {'tree': treeFullPath, 'ext': extension}
            )
        LOG.debug(debugMessage)

      else:
        debugMessage = "Sandboxing : Tree '%s' already exists" % treeFullPath
        LOG.debug(debugMessage)

      # We take the first directory of the tree
      # We add this root directory to the 'final' sandbox
      rootFolder = tree.split(os.path.sep)[0]
      rootFolderFullPath = os.path.join(self._tempCwd, rootFolder)

      self._inputSandbox.add(rootFolderFullPath)

      # if paths already exists do not copy it
      # go to the next file
      if os.path.exists(destination):
        debugMessage = "Sandboxing : The file '%s' already exists" % destination
        LOG.debug(debugMessage)
      else:
        debugMessage = "Sandboxing : Additional file '%s' copy..." % source
        LOG.debug(debugMessage)

        try:
          shutil.copyfile(source, destination)
        except (IOError, shutil.Error) as e:
          errorMessage = (
              "Sandboxing : Additionnal files"
              " '%(src)s' copy failed\n%(error)s" % {'src': source, 'error': e}
              )
          LOG.error(errorMessage)
          return False

        debugMessage = (
            "Sandboxing : Additionnal files"
            " '%(src)s' copy successfull to '%(dst)s'" % {'src': source, 'dst': destination}
            )
        LOG.debug(debugMessage)

    return True

  def _setFilterToFolders(self):
    """Some folders required by FCCSW do not need to be imported with all their content.

    Some files have to be excluded or only some files have to be included.
    Then for each folder, we have the include/exclude parameter and the linked extension.

    This function maps the folders with their corresponding filters if there are.

    :return: The success or the failure of the filter's setting
    :rtype: bool
    """

    if not self._foldersToFilter:
      debugMessage = "Sandboxing : No filtering required"
      LOG.debug(debugMessage)
      return True

    copiedFolders = []

    for idx, actualFolder in enumerate(self._foldersToFilter):

      if not os.path.exists(actualFolder):
        errorMessage = (
            "Sandboxing : _filterFolders() failed\n"
            "The folder '%(actual)s' does not exist\n"
            "Check if you're FCCSW installation is complete" % {'actual': actualFolder}
            )
        LOG.error(errorMessage)
        return False

      if idx < len(self._filteredExtensions):
        filteredExtension = self._filteredExtensions[idx]
        excludeOrInclude = self._excludesOrIncludes[idx]
      else:
        filteredExtension = False
        excludeOrInclude = False

      tempFolder = os.path.join(self._tempCwd, os.path.basename(actualFolder))

      # DIRAC already compress the sandbox before submitting the job
      # do not compress folders

      LOG.debug("Sandboxing : Folders filtering...")

      if not self._filterFolders(tempFolder, actualFolder, filteredExtension, excludeOrInclude):
        LOG.error("Sandboxing : _filterFolders() failed")
        return False

      LOG.debug("Sandboxing : Folders filtering successfull")

      copiedFolders.append(tempFolder)

    self._foldersToFilter = copiedFolders

    return True

  def _filterFolders(self, tempFolder, actualFolder,
                     filteredExtension, excludeOrInclude):
    """Knowing the filter for each folder, we copy the 'filtered' folder in the temporary folder 'temp_fcc_dirac'.

    After, we add the path of the 'filtered' folder to the sandbox instead of
    its original path.

    Like that, we do not import unnecessary files to the sandbox.
    It is like we create a 'light' copy of the orginal folder.

    If folders have to be filtered then, we do not have to add them directly to the sandbox
    because it will put all content of folders.

    So we copy only the content we want in the filtered folder
    inside the temporary folder 'temp_fcc_dirac'.

    After, we add the filtered folders located in 'temp_fcc_dirac' to the sandbox.

    :param tempFolder: The temporary working directory (destination) used for the sandboxing
    :type tempFolder: str

    :param actualFolder: The original (source) path of folder checked by the copy process
    :type actualFolder: str

    :param filteredExtension: The extension of file we do (not) want
    :type filteredExtension: str

    :param excludeOrInclude: extension is excluded or included
    :type excludeOrInclude: bool

    :return: The success or the failure of the filtering
    :rtype: bool
    """

    debugMessage = "Sandboxing : Checking of the filtered folder '%s'..." % tempFolder
    LOG.debug(debugMessage)

    if not os.path.exists(tempFolder):
      debugMessage = "Sandboxing : Creation of the filtered folder '%s'..." % tempFolder
      LOG.debug(debugMessage)

      try:
        os.makedirs(tempFolder)
      except OSError as e:
        errorMessage = (
            "Sandboxing : Creation of the filtered folder"
            " '%(temp)s' failed\n%(error)s" % {'temp': tempFolder, 'error': e}
            )
        LOG.error(errorMessage)
        return False

      debugMessage = (
          "Sandboxing : Creation of the filtered folder"
          " '%(temp)s' successfull" % {'temp': tempFolder}
          )
      LOG.debug(debugMessage)

    for path in os.listdir(actualFolder):
      source = os.path.realpath(os.path.join(actualFolder, path))
      destination = os.path.realpath(os.path.join(tempFolder, path))

      # If file then check existence
      if not os.path.exists(source):
        errorMessage = "Sandboxing : The file '%s' does not exist" % source
        LOG.error(errorMessage)
        return False

      if not os.path.isfile(source):
        # Recursive call for folders
        if not self._filterFolders(destination, source, filteredExtension, excludeOrInclude):
          LOG.error("Sandboxing : _filterFolders() failed")
          return False
      else:
        if os.path.exists(destination):
          debugMessage = "Sandboxing : The file '%s' already exists" % source
          LOG.debug(debugMessage)
        else:

          if ((excludeOrInclude and not path.endswith(filteredExtension))
              or (not excludeOrInclude and path.endswith(filteredExtension))
                  or not filteredExtension):

            warn = False
            debugMessage = "Sandboxing : File '%s' copy..." % source
            LOG.debug(debugMessage)

            # Copy considering filters to apply
            try:
              shutil.copyfile(source, destination)
            except (IOError, shutil.Error) as e:
              warnMessage = "Sandboxing : The copy of the file '%s' failed\n%s" % (destination, e)
              LOG.warn(warnMessage)
              warn = True
              # return False

            if not warn:
              debugMessage = (
                  "Sandboxing : Copy of the file"
                  " '%(src)s' successfull to '%(dst)s'" % {'src': source, 'dst': destination}
                  )
              LOG.debug(debugMessage)

    debugMessage = "Sandboxing : Folder '%s' filtering successfull" % tempFolder
    LOG.debug(debugMessage)

    return True


class FccAnalysis(Fcc):
  """Definition of an FCCAnalysis application. By default, it runs FCCPHYSICS.

  Usage:

  >>> FCC_PHYSICS = FccAnalysis(
      fccConfFile='/cvmfs/fcc.cern.ch/sw/0.7/fcc-physics/0.1/x86_64-slc6-gcc49-opt/share/ee_ZH_Zmumu_Hbb.txt'
    )
  """

  def __init__(self, executable='fcc-pythia8-generate', fccConfFile=""):

    super(FccAnalysis, self).__init__()

    self.steeringFile = fccConfFile
    self.fccExecutable = executable

    # If it is a different executable
    # then it is :
    # - fcc-physics-read-delphes or
    # - fcc-physics-read
    # So find card file and set the seed !

    if executable != 'fcc-pythia8-generate':
      self.read = True
    else:  # Set the seed for the generation of events
      self.randomGenerator = {"Pythia": []}



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Test FileStatusTransformationAgent."""

from __future__ import absolute_import
import unittest
import sys
from contextlib import contextmanager
from mock import MagicMock, patch

import ILCDIRAC.ILCTransformationSystem.Agent.FileStatusTransformationAgent as FST
from ILCDIRAC.ILCTransformationSystem.Agent.FileStatusTransformationAgent import FileStatusTransformationAgent
import DIRAC.Resources.Storage.StorageElement as SeModule

from DIRAC import S_OK, S_ERROR, gLogger
import DIRAC

__RCSID__ = "$Id$"


@contextmanager
def mockedUserProxy(*_args, **_kwargs):
  """Context manager to replace the UserProxy context manager."""
  try:
    yield S_OK()
  finally:
    pass


@contextmanager
def mockedUserProxyError(*_args, **_kwargs):
  """Context manager to replace the UserProxy context manager."""
  try:
    yield S_ERROR()
  finally:
    pass


class TestFSTAgent(unittest.TestCase):
  """Test_fstAgent class."""

  def setUp(self):
    self.agent = FST
    self.agent.AgentModule = MagicMock()
    self.fstAgent = FileStatusTransformationAgent()

    self.fstAgent.log = gLogger
    self.fstAgent.tClient = MagicMock(
        name="transMock", spec=DIRAC.TransformationSystem.Client.TransformationClient.TransformationClient)
    self.fstAgent.fcClient = MagicMock(name="fcMock", spec=DIRAC.Resources.Catalog.FileCatalogClient.FileCatalogClient)
    self.fstAgent.reqClient = MagicMock(name="reqMock", spec=DIRAC.RequestManagementSystem.Client.ReqClient)
    self.fstAgent.nClient = MagicMock(name="nMock", spec=DIRAC.FrameworkSystem.Client.NotificationClient)

    self.fstAgent.tClient.setTaskStatus = MagicMock()
    self.fstAgent.enabled = True

    self.fakeTransID = 1

    self.failedTask = {'TargetSE': 'DESY-SRM',
                       'TransformationID': 400103,
                       'ExternalStatus': 'Failed',
                       'ExternalID': 0,
                       'TaskID': 0}

    self.doneTask = {'TargetSE': 'DESY-SRM',
                     'TransformationID': 400103,
                     'ExternalID': 1,
                     'ExternalStatus': 'Done',
                     'TaskID': 1}

    self.notAvailableOnSrc = '/ilc/file_not_available_on_src'
    self.notAvailableOnDst = '/ilc/file_not_available_on_target'
    self.available = '/ilc/file_on_src_and_target'
    self.notAvailable = '/ilc/file_not_on_src_and_target'

    self.sourceSE = ['CERN-SRM']
    self.targetSE = ['DESY-SRM']

  @classmethod
  def tearDownClass(cls):
    sys.modules.pop('ILCDIRAC.ILCTransformationSystem.Agent.FileStatusTransformationAgent')

  def test_init(self):
    self.assertIsInstance(self.fstAgent, FileStatusTransformationAgent)
    self.assertIsInstance(self.fstAgent.tClient, MagicMock)
    self.assertIsInstance(self.fstAgent.fcClient, MagicMock)
    self.assertIsInstance(self.fstAgent.reqClient, MagicMock)
    self.assertTrue(self.fstAgent.enabled)
    self.assertEqual(self.fstAgent.transformationTypes, ['Replication'])
    self.assertEqual(self.fstAgent.transformationStatuses, ['Active'])

  def test_get_transformations_failure(self):
    """fstAgent should stop execution cycle if tClient getTransformations returns an error."""
    self.fstAgent.processTransformation = MagicMock()
    self.fstAgent.tClient.getTransformations.return_value = S_ERROR()

    res = self.fstAgent.execute()
    self.fstAgent.processTransformation.assert_not_called()
    self.assertFalse(res['OK'])

  def test_no_transformations_found(self):
    """fstAgent should stop execution cycle if no transformations are found."""
    self.fstAgent.processTransformation = MagicMock()
    self.fstAgent.tClient.getTransformations.return_value = S_OK([])

    res = self.fstAgent.execute()
    self.fstAgent.processTransformation.assert_not_called()
    self.assertTrue(res['OK'])

  def test_process_transformation_failure(self):
    """fstAgent should not exit if processing of some transformation returns an Error, all transformations should be processed independently."""
    self.fstAgent.processTransformation = MagicMock()
    self.fstAgent.getTransformations = MagicMock()

    transformations = [{'TransformationID': 1, 'SourceSE': ['CERN'], 'TargetSE': ['DESY'], 'DataTransType': 'x'},
                       {'TransformationID': 2, 'SourceSE': ['CERN'], 'TargetSE': ['DESY'], 'DataTransType': 'x'}]
    self.fstAgent.getTransformations.return_value = S_OK(transformations)
    self.fstAgent.processTransformation.return_value = S_ERROR()

    self.fstAgent.execute()
    assert len(self.fstAgent.processTransformation.mock_calls) == 2

  def _getOption(self, option, defaultVal):
    if option != "TransformationFileStatuses":
      return defaultVal
    defaultVal.append("UnknownStatus")
    return defaultVal

  def test_unknown_file_status(self):
    """fstAgent should not process file statuses for whom check files function is not implemented."""
    allowedFileStatuses = ["Assigned", "Problematic", "Processed", "Unused"]
    self.fstAgent.am_getOption = MagicMock(side_effect=self._getOption)
    self.fstAgent.am_setOption = MagicMock(side_effect=self._getOption)
    self.fstAgent.beginExecution()
    assert sorted(self.fstAgent.transformationFileStatuses) == sorted(allowedFileStatuses)

  def test_get_transformations(self):
    """Test for getTransformations function."""
    self.fstAgent.tClient.getTransformationParameters = MagicMock()
    self.fstAgent.getDataTransformationType = MagicMock()
    self.fstAgent.processTransformation = MagicMock()
    self.fstAgent.sendNotification = MagicMock()

    self.fstAgent.tClient.getTransformations.return_value = S_ERROR()
    res = self.fstAgent.getTransformations()
    self.assertFalse(res['OK'])
    self.fstAgent.tClient.getTransformations.reset_mock()
    res = self.fstAgent.getTransformations(transID=self.fakeTransID)
    self.fstAgent.tClient.getTransformations.assert_called_once_with(
        condDict={'TransformationID': self.fakeTransID,
                  'Status': self.fstAgent.transformationStatuses,
                  'Type': self.fstAgent.transformationTypes})

    transInfoDict = {'Status': 'Active', 'TransformationID': self.fakeTransID, 'Type': 'Replication'}
    self.fstAgent.tClient.getTransformations.return_value = S_OK([transInfoDict])
    self.fstAgent.tClient.getTransformationParameters.return_value = S_ERROR()
    self.fstAgent.execute()
    self.fstAgent.processTransformation.assert_not_called()
    self.fstAgent.sendNotification.assert_called()

    self.fstAgent.tClient.getTransformationParameters.return_value = S_OK({'TargetSE': "['CERN-DST-EOS']",
                                                                           'SourceSE': "['CERN-SRM']"})
    self.fstAgent.processTransformation.reset_mock()
    self.fstAgent.getDataTransformationType.return_value = S_ERROR()
    self.fstAgent.execute()
    self.fstAgent.processTransformation.assert_not_called()
    self.fstAgent.sendNotification.assert_called()

    self.fstAgent.getDataTransformationType.return_value = S_OK(FST.REPLICATION_TRANS)
    self.fstAgent.processTransformation.reset_mock()
    self.fstAgent.execute()
    self.fstAgent.processTransformation.assert_called_once_with(self.fakeTransID, ['CERN-SRM'], ['CERN-DST-EOS'],
                                                                FST.REPLICATION_TRANS, transInfoDict)

  def test_send_notification(self):
    """Test for sendNotification function."""
    dataTransType = FST.REPLICATION_TRANS
    sourceSE = ['CERN-SRM']
    targetSE = ['DESY-SRM']
    errList = ["some error occured", "some other error"]
    accDict = {FST.SET_PROCESSED: [{'LFN': '/ilc/fake/lfn',
                                    'Status': 'Problematic',
                                    'AvailableOnSource': True,
                                    'AvailableOnTarget': True}]}
    self.fstAgent.nClient.sendMail = MagicMock()

    # email should not be sent if accounting dict and errors list is empty
    self.fstAgent.accounting = {}
    self.fstAgent.errors = []
    self.fstAgent.sendNotification(self.fakeTransID, dataTransType, sourceSE, targetSE)
    self.fstAgent.nClient.sendMail.assert_not_called()

    # email should be sent if accounting dict is empty but errors list contains some error strings
    self.fstAgent.errors = errList
    self.fstAgent.sendNotification(self.fakeTransID, dataTransType, sourceSE, targetSE)
    self.fstAgent.nClient.sendMail.assert_called()

    # email should be sent if errors list is empty but accounting dict contains some values
    self.fstAgent.nClient.sendMail.reset_mock()
    self.fstAgent.accounting = accDict
    self.fstAgent.errors = []
    self.fstAgent.sendNotification(self.fakeTransID, dataTransType, sourceSE, targetSE)
    self.fstAgent.nClient.sendMail.assert_called()

    # try sending email to all addresses even if we get error for sending email to some address
    self.fstAgent.nClient.sendMail.reset_mock()
    self.fstAgent.errors = errList
    self.fstAgent.accounting = accDict
    self.fstAgent.addressTo = ["name1@cern.ch", "name2@cern.ch"]
    self.fstAgent.nClient.sendMail.return_value = S_ERROR()
    res = self.fstAgent.sendNotification(self.fakeTransID, dataTransType, sourceSE, targetSE)
    self.assertEqual(len(self.fstAgent.nClient.sendMail.mock_calls), len(self.fstAgent.addressTo))
    self.assertTrue(res['OK'])

    # accounting dict and errors list should be cleared after notification is sent
    self.assertEqual(self.fstAgent.accounting, {})
    self.assertEqual(self.fstAgent.errors, [])

  def test_get_data_transformation_type(self):
    """Test if getDataTransformationType function correctly returns the Data Transformation Type (Replication / Moving)"""
    self.fstAgent.tClient.getTransformationParameters = MagicMock()

    self.fstAgent.tClient.getTransformationParameters.return_value = S_ERROR()
    res = self.fstAgent.getDataTransformationType(self.fakeTransID)
    self.assertFalse(res['OK'])

    self.fstAgent.tClient.getTransformationParameters.return_value = S_OK("")
    res = self.fstAgent.getDataTransformationType(self.fakeTransID)['Value']
    self.assertEqual(res, FST.REPLICATION_TRANS)

    self.fstAgent.tClient.getTransformationParameters.return_value = S_OK("sdfdsfdsfs")
    res = self.fstAgent.getDataTransformationType(self.fakeTransID)
    self.assertFalse(res['OK'])

    self.fstAgent.tClient.getTransformationParameters.return_value = S_OK(
        '[["ReplicateAndRegister", {"TargetSE": ["CERN-SRM"], "SourceSE": "CERN-DST-EOS"}],'
        '["RemoveReplica", {"TargetSE": "CERN-DST-EOS"}]]')
    res = self.fstAgent.getDataTransformationType(self.fakeTransID)['Value']
    self.assertEqual(res, FST.MOVING_TRANS)

    self.fstAgent.tClient.getTransformationParameters.return_value = S_OK(
        '[["RemoveReplica", {"TargetSE": "CERN-DST-EOS"}]]')
    res = self.fstAgent.getDataTransformationType(self.fakeTransID)
    self.assertFalse(res['OK'])

    self.fstAgent.tClient.getTransformationParameters.return_value = S_OK(
        'RemoveReplica:CERN-DST-EOS;ReplicateAndRegister')
    res = self.fstAgent.getDataTransformationType(self.fakeTransID)['Value']
    self.assertEqual(res, FST.MOVING_TRANS)

    self.fstAgent.tClient.getTransformationParameters.return_value = S_OK('RemoveReplica:CERN-DST-EOS')
    res = self.fstAgent.getDataTransformationType(self.fakeTransID)
    self.assertFalse(res['OK'])

    self.fstAgent.tClient.getTransformationParameters.return_value = S_OK(
        '[["ReplicateAndRegister", {"TargetSE": ["CERN-SRM"], "SourceSE": "CERN-DST-EOS"}]]')
    res = self.fstAgent.getDataTransformationType(self.fakeTransID)['Value']
    self.assertEqual(res, FST.REPLICATION_TRANS)

    self.fstAgent.tClient.getTransformationParameters.return_value = S_OK('ReplicateAndRegister')
    res = self.fstAgent.getDataTransformationType(self.fakeTransID)['Value']
    self.assertEqual(res, FST.REPLICATION_TRANS)

  def test_get_request_status(self):
    """Test getRequestStatus function."""
    taskID = 1
    taskIDs = [taskID]
    self.fstAgent.tClient.getTransformationTasks.return_value = S_ERROR()
    res = self.fstAgent.getRequestStatus(self.fakeTransID, taskIDs)
    self.assertFalse(res['OK'])

    self.fstAgent.tClient.getTransformationTasks.return_value = S_OK([{'TaskID': taskID,
                                                                       'ExternalStatus': 'Failed',
                                                                       'ExternalID': 123}])
    res = self.fstAgent.getRequestStatus(self.fakeTransID, taskIDs)
    result = res['Value']
    self.assertEqual(result[taskID]['RequestStatus'], 'Failed')
    self.assertEqual(result[taskID]['RequestID'], 123)

  def test_set_file_status(self):
    """Test for setFileStatus function."""
    transFiles = []
    newStatus = "processed"
    self.fstAgent.setFileStatus(self.fakeTransID, transFiles, newStatus)
    self.fstAgent.tClient.setFileStatusForTransformation.assert_not_called()

    lfn1 = '/ilc/fake/lfn1'
    lfn2 = '/ilc/fake/lfn2'
    transFiles = [{'LFN': lfn1, 'Status': 'Problematic', 'AvailableOnSource': True, 'AvailableOnTarget': True},
                  {'LFN': lfn2, 'Status': 'Problematic', 'AvailableOnSource': True, 'AvailableOnTarget': True}]

    self.fstAgent.tClient.setFileStatusForTransformation.return_value = S_ERROR()
    res = self.fstAgent.setFileStatus(self.fakeTransID, transFiles, newStatus)
    self.assertFalse(res['OK'])

    self.fstAgent.tClient.setFileStatusForTransformation.reset_mock()
    self.fstAgent.tClient.setFileStatusForTransformation.return_value = S_OK()
    self.fstAgent.setFileStatus(self.fakeTransID, transFiles, newStatus)
    self.fstAgent.tClient.setFileStatusForTransformation.assert_called_once_with(self.fakeTransID, newLFNsStatus={
                                                                                 lfn1: newStatus, lfn2: newStatus},
                                                                                 force=True)
    self.assertTrue(newStatus in self.fstAgent.accounting)
    self.assertEqual(len(self.fstAgent.accounting[newStatus]), 2)

  def test_exists_in_FC(self):
    """Test if the existsInFC function correctly determines if all replicas of files are registered in FC."""
    se1 = 'CERN-SRM'
    se2 = 'DESY-SRM'
    storageElements = [se1, se2]

    # file exists in FC and all replicas are registered
    fileExists = '/ilc/file/file1'

    # file exists in FC, but one replica is missing in FC
    fileOneRepLost = '/ilc/file/file2'

    # file exists in FC, but no replicas are registered in FC
    fileAllRepLost = '/ilc/file/file3'

    # file does not exist in FC
    fileRemoved = '/ilc/file/file4'

    files = [fileExists, fileOneRepLost, fileAllRepLost, fileRemoved]

    self.fstAgent.fcClient.getReplicas.return_value = S_ERROR()
    res = self.fstAgent.existsInFC(storageElements, files)
    self.assertFalse(res['OK'])

    self.fstAgent.fcClient.getReplicas.return_value = S_OK({'Successful': {fileExists: {se1: fileExists,
                                                                                        se2: fileExists},
                                                                           fileOneRepLost: {se1: fileOneRepLost},
                                                                           fileAllRepLost: {}},
                                                            'Failed': {fileRemoved: 'No such file or directory'}})
    res = self.fstAgent.existsInFC(storageElements, files)['Value']
    self.assertTrue(res['Successful'][fileExists])
    self.assertFalse(res['Successful'][fileOneRepLost])
    self.assertFalse(res['Successful'][fileAllRepLost])
    self.assertFalse(res['Successful'][fileRemoved])

  @patch('ILCDIRAC.ILCTransformationSystem.Agent.FileStatusTransformationAgent.UserProxy', new=mockedUserProxy)
  def test_exists_on_storage_element(self):
    """Test if the existsOnSE function correctly determines if a file exists on all provided Storage Elements or not."""

    se1 = 'CERN-SRM'
    se2 = 'DESY-SRM'
    storageElements = [se1, se2]

    # file exists on all SEs
    fileExists = '/ilc/file/file1'

    # file is lost on one SE
    fileOneRepLost = '/ilc/file/file2'

    # file is lost on all SEs
    fileAllRepLost = '/ilc/file/file3'

    # some error to get file status on SE
    fileFailed = '/ilc/file/file4'

    authorInfoDict = {'AuthorDN': 'A_DN', 'AuthorGroup': 'A_GROUP'}

    files = [fileExists, fileOneRepLost, fileAllRepLost, fileFailed]

    se1Result = S_OK({'Successful': {fileExists: True, fileOneRepLost: True, fileAllRepLost: False, fileFailed: True},
                      'Failed': {}})

    se2Result = S_OK({'Successful': {fileExists: True, fileOneRepLost: False, fileAllRepLost: False},
                      'Failed': {fileFailed: 'permission denied'}})

    SeModule.StorageElementItem.exists = MagicMock()
    SeModule.StorageElementItem.exists.return_value = S_ERROR()
    res = self.fstAgent.existsOnSE(storageElements, files, authorInfoDict)
    self.assertFalse(res['OK'])

    res = self.fstAgent.existsOnSE(storageElements, [], authorInfoDict)
    self.assertTrue(res['OK'])
    self.assertEqual(res['Value']['Successful'], {})
    self.assertEqual(res['Value']['Failed'], {})

    SeModule.StorageElementItem.exists.side_effect = [se1Result, se2Result]
    res = self.fstAgent.existsOnSE(storageElements, files, authorInfoDict)['Value']

    self.assertTrue(res['Successful'][fileExists])
    self.assertFalse(res['Successful'][fileAllRepLost])
    self.assertFalse(res['Successful'][fileOneRepLost])
    self.assertTrue(len(res['Failed']), 1)

  def test_exists_on_storage_element_failProxy(self):
    """Test when failing to get a proxy."""
    with patch('ILCDIRAC.ILCTransformationSystem.Agent.FileStatusTransformationAgent.UserProxy',
               new=mockedUserProxyError):
      authorInfoDict = {'AuthorDN': 'A_DN', 'AuthorGroup': 'A_GROUP'}
      res = self.fstAgent.existsOnSE([], ['/vo/lfns'], authorInfoDict)
      self.assertFalse(res['OK'])

  def test_exists(self):
    """Tests if the exists function correctly determines if a file exists in File Catalog and Storage Elements."""

    se1 = 'CERN-SRM'
    se2 = 'DESY-SRM'
    storageElements = ['CERN-SRM', 'DESY-SRM']

    # file exists in fc and all SEs
    fileExists = '/ilc/file/file1'

    # file exists in fc but one replica is lost on SE
    fileOneRepLostOnSE = '/ilc/file/file2'

    # file exists in fc but all replicas are lost on SEs
    fileAllRepLostOnSEs = '/ilc/file/file3'

    # file does not exists in FC
    fileRemoved = '/ilc/file/file4'

    transInfoDict = [{'TransformationID': 1, 'SourceSE': ['CERN'], 'TargetSE': ['DESY'], 'DataTransType': 'x',
                      'AuthorDN': 'A_DN', 'AuthorGroup': 'A_GROUP'}]

    files = [fileExists, fileOneRepLostOnSE, fileAllRepLostOnSEs, fileRemoved]

    self.fstAgent.existsInFC = MagicMock()
    self.fstAgent.existsOnSE = MagicMock()

    self.fstAgent.existsInFC.return_value = S_ERROR()

    res = self.fstAgent.exists(storageElements, files, transInfoDict)
    self.assertFalse(res['OK'])

    # if no files were found in FC, no need to check the lfns on storage
    self.fstAgent.existsInFC.return_value = S_OK({'Successful': {fileRemoved: False},
                                                  'Failed': {}})
    self.fstAgent.exists(storageElements, [fileRemoved], transInfoDict)
    self.fstAgent.existsOnSE.assert_not_called()

    self.fstAgent.existsInFC.return_value = S_OK({'Successful': {fileExists: True, fileOneRepLostOnSE: True,
                                                                 fileAllRepLostOnSEs: True, fileRemoved: False},
                                                  'Failed': {}})
    self.fstAgent.existsOnSE.return_value = S_ERROR()
    res = self.fstAgent.exists(storageElements, files, transInfoDict)
    self.assertFalse(res['OK'])

    self.fstAgent.existsOnSE.return_value = S_OK({'Successful': {},
                                                  'Failed': {se1: {fileExists: "permission denied",
                                                                   fileOneRepLostOnSE: "permission denied",
                                                                   fileAllRepLostOnSEs: "permission denied"},
                                                             se2: {}}})
    res = self.fstAgent.exists(storageElements, files, transInfoDict)
    self.assertFalse(res['OK'])

    self.fstAgent.existsOnSE.return_value = S_OK({'Successful': {fileExists: True, fileOneRepLostOnSE: False,
                                                                 fileAllRepLostOnSEs: False},
                                                  'Failed': {se1: {}, se2: {}}})
    res = self.fstAgent.exists(storageElements, files, transInfoDict)['Value']['Successful']
    self.assertTrue(res[fileExists])
    self.assertFalse(res[fileOneRepLostOnSE])
    self.assertFalse(res[fileAllRepLostOnSEs])
    self.assertFalse(res[fileRemoved])

  def test_select_failed_requests(self):
    """Test if selectFailedRequests function returns True if transfile has a failed request."""

    transFileWithFailedReq = {'TransformationID': 400103, 'TaskID': 0, 'LFN': '/ilc/file1'}
    transFileWithDoneReq = {'TransformationID': 400103, 'TaskID': 1, 'LFN': '/ilc/file2'}

    self.fstAgent.tClient.getTransformationTasks.return_value = S_ERROR()
    res = self.fstAgent.selectFailedRequests(transFileWithFailedReq)
    self.assertFalse(res)

    self.fstAgent.tClient.getTransformationTasks.return_value = S_OK([self.failedTask])
    res = self.fstAgent.selectFailedRequests(transFileWithFailedReq)
    self.assertTrue(res)

    self.fstAgent.tClient.getTransformationTasks.return_value = S_OK([self.doneTask])
    res = self.fstAgent.selectFailedRequests(transFileWithDoneReq)
    self.assertFalse(res)

  def test_retry_strategy_for_files(self):
    """Test if the request exists then retry strategy is resetting the request otherwise set the file to unused."""

    taskIDfile1 = 1
    taskIDfile2 = 2
    transFiles = [{'TransformationID': self.fakeTransID, 'TaskID': taskIDfile1, 'LFN': '/ilc/file1'},
                  {'TransformationID': self.fakeTransID, 'TaskID': taskIDfile2, 'LFN': '/ilc/file2'}]

    self.fstAgent.getRequestStatus = MagicMock()
    self.fstAgent.getRequestStatus.return_value = S_ERROR()
    res = self.fstAgent.retryStrategyForFiles(self.fakeTransID, transFiles)
    self.assertFalse(res['OK'])

    self.fstAgent.getRequestStatus.return_value = S_OK({taskIDfile1: {'RequestStatus': 'Problematic',
                                                                      'RequestID': 1},
                                                        taskIDfile2: {'RequestStatus': 'Problematic',
                                                                      'RequestID': 2}})

    # no request exists for first trans file and one request exists for second trans file
    self.fstAgent.reqClient.getRequest = MagicMock()
    self.fstAgent.reqClient.getRequest.side_effect = [S_ERROR('Request does not exist'), S_OK('Request exists')]

    res = self.fstAgent.retryStrategyForFiles(self.fakeTransID, transFiles)['Value']

    self.assertEqual(res[taskIDfile1]['Strategy'], FST.SET_UNUSED)
    self.assertEqual(res[taskIDfile2]['Strategy'], FST.SET_UNUSED)

  def test_retry_files(self):
    """Test for retryFiles function."""
    transFiles = [{'LFN': '/ilc/file1', 'Status': 'Problematic', 'AvailableOnSource': True,
                   'AvailableOnTarget': True, 'TaskID': 1},
                  {'LFN': '/ilc/file2', 'Status': 'Problematic', 'AvailableOnSource': True,
                   'AvailableOnTarget': True, 'TaskID': 2}]

    self.fstAgent.retryStrategyForFiles = MagicMock()
    self.fstAgent.setFileStatus = MagicMock()
    self.fstAgent.reqClient.resetFailedRequest = MagicMock()

    self.fstAgent.retryStrategyForFiles.return_value = S_ERROR()
    res = self.fstAgent.retryFiles(self.fakeTransID, transFiles)
    self.assertFalse(res['OK'])

    self.fstAgent.retryStrategyForFiles.return_value = S_OK({1: {'Strategy': FST.RESET_REQUEST,
                                                                 'RequestID': 1},
                                                             2: {'Strategy': FST.SET_UNUSED}})
    self.fstAgent.reqClient.resetFailedRequest.return_value = S_ERROR()
    self.fstAgent.retryFiles(self.fakeTransID, transFiles)
    self.fstAgent.tClient.setTaskStatus.assert_not_called()

    self.fstAgent.reqClient.resetFailedRequest.return_value = S_OK("Not reset")
    self.fstAgent.retryFiles(self.fakeTransID, transFiles)
    self.fstAgent.tClient.setTaskStatus.assert_not_called()

    self.fstAgent.tClient.setTaskStatus.reset_mock()
    self.fstAgent.reqClient.resetFailedRequest.reset_mock()
    self.fstAgent.setFileStatus.reset_mock()
    self.fstAgent.reqClient.resetFailedRequest.return_value = S_OK()
    self.fstAgent.tClient.setTaskStatus.return_value = S_OK()
    self.fstAgent.retryFiles(self.fakeTransID, transFiles)
    self.fstAgent.reqClient.resetFailedRequest.assert_called_once_with(1, allR=True)
    self.fstAgent.tClient.setTaskStatus.assert_called_once_with(self.fakeTransID, 1, 'Waiting')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [transFiles[1]], 'Unused')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [transFiles[0]], 'Assigned')

    # set file statues to unused should not be called if all files have reset request strategy
    self.fstAgent.retryStrategyForFiles.return_value = S_OK({1: {'Strategy': FST.RESET_REQUEST,
                                                                 'RequestID': 1},
                                                             2: {'Strategy': FST.RESET_REQUEST,
                                                                 'RequestID': 2}})
    self.fstAgent.setFileStatus.reset_mock()
    self.fstAgent.reqClient.resetFailedRequest.reset_mock()
    self.fstAgent.retryFiles(self.fakeTransID, transFiles)
    self.assertEqual(len(self.fstAgent.reqClient.resetFailedRequest.mock_calls), 2)
    self.fstAgent.setFileStatus.assert_called_once_with(self.fakeTransID, transFiles, 'Assigned')

  def _exists(self, se, lfns, _transInfoDict):
    """returns lfns availability information."""
    result = {}
    for lfn in lfns:
      if (lfn == self.notAvailableOnSrc or lfn == self.notAvailable) and se == self.sourceSE:
        result[lfn] = False
      elif (lfn == self.notAvailableOnDst or lfn == self.notAvailable) and se == self.targetSE:
        result[lfn] = False
      else:
        result[lfn] = True

    return S_OK({'Successful': result})

  def test_trans_files_treatment(self):
    """test transformation files are treated properly (set new status / reset request) for replication and moving transformations."""
    self.fstAgent.setFileStatus = MagicMock()
    self.fstAgent.sendNotification = MagicMock()

    fileNotAvailableOnSrc = {'TransformationID': self.fakeTransID, 'TaskID': 1, 'LFN': self.notAvailableOnSrc}
    fileNotAvailableOnDst = {'TransformationID': self.fakeTransID, 'TaskID': 2, 'LFN': self.notAvailableOnDst}
    fileAvailable = {'TransformationID': self.fakeTransID, 'TaskID': 3, 'LFN': self.available}
    fileNotAvailable = {'TransformationID': self.fakeTransID, 'TaskID': 4, 'LFN': self.notAvailable}

    authorInfoDict = {'AuthorDN': 'A_DN', 'AuthorGroup': 'A_GROUP'}

    transFiles = [fileNotAvailableOnSrc, fileNotAvailableOnDst, fileAvailable, fileNotAvailable]

    # all trans files have failed requests
    self.fstAgent.selectFailedRequests = MagicMock(return_value={tFile['TaskID']: True for tFile in transFiles})

    # no assosiated request in rms
    self.fstAgent.retryStrategyForFiles = MagicMock(return_value=S_OK(
        {tFile['TaskID']: {'Strategy': FST.SET_UNUSED} for tFile in transFiles}))

    self.fstAgent.exists = MagicMock()

    # all file statuses should be processed even if getTransformationFiles returns an error for some status
    self.fstAgent.transformationFileStatuses = ['Assigned', 'Problematic']
    self.fstAgent.tClient.getTransformationFiles.return_value = S_ERROR()
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.REPLICATION_TRANS,
                                        authorInfoDict)
    self.assertEqual(len(self.fstAgent.tClient.getTransformationFiles.mock_calls), 2)

    # all file statuses should be processed even if no transformation files are found for some status
    self.fstAgent.tClient.getTransformationFiles.reset_mock()
    self.fstAgent.tClient.getTransformationFiles.return_value = S_OK([])
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.REPLICATION_TRANS,
                                        authorInfoDict)
    self.assertEqual(len(self.fstAgent.tClient.getTransformationFiles.mock_calls), 2)

    # all file statuses should be processed even if we get a failure to determine if transformation files
    # with some status exists in FileCatalog and StorageElements
    self.fstAgent.tClient.getTransformationFiles.reset_mock()
    self.fstAgent.tClient.getTransformationFiles.return_value = S_OK(transFiles)
    self.fstAgent.exists.return_value = S_ERROR()
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.REPLICATION_TRANS,
                                        authorInfoDict)
    self.assertEqual(len(self.fstAgent.tClient.getTransformationFiles.mock_calls), 2)

    self.fstAgent.tClient.getTransformationFiles.return_value = S_OK(transFiles)
    self.fstAgent.exists.side_effect = self._exists
    self.fstAgent.transformationFileStatuses = ['Assigned']

    # check replication transformation treatment for assigned files
    self.fstAgent.setFileStatus.reset_mock()
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.REPLICATION_TRANS,
                                        authorInfoDict)
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnSrc, fileAvailable], 'Processed')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnDst], 'Unused')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailable], 'Deleted')

    # check moving transformation treatment for assigned files
    self.fstAgent.setFileStatus.reset_mock()
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.MOVING_TRANS,
                                        authorInfoDict)
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnSrc], 'Processed')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnDst, fileAvailable], 'Unused')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailable], 'Deleted')

    self.fstAgent.transformationFileStatuses = ['Processed']

    # check replication transformation treatment for processed files
    self.fstAgent.setFileStatus.reset_mock()
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.REPLICATION_TRANS,
                                        authorInfoDict)
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnDst], 'Unused')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailable], 'Deleted')

    # check moving transformation treatment for processed files
    self.fstAgent.setFileStatus.reset_mock()
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.MOVING_TRANS,
                                        authorInfoDict)
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnDst, fileAvailable], 'Unused')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailable], 'Deleted')

    self.fstAgent.transformationFileStatuses = ['Problematic']

    # check replication transformation treatment for problematic files
    self.fstAgent.setFileStatus.reset_mock()
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.REPLICATION_TRANS,
                                        authorInfoDict)
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnSrc, fileAvailable], 'Processed')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnDst], 'Unused')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailable], 'Deleted')

    # check moving transformation treatment for problematic files
    self.fstAgent.setFileStatus.reset_mock()
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.MOVING_TRANS,
                                        authorInfoDict)
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnSrc], 'Processed')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnDst, fileAvailable], 'Unused')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailable], 'Deleted')

    self.fstAgent.transformationFileStatuses = ['Unused']

    # check replication transformation treatment for unused files
    self.fstAgent.setFileStatus.reset_mock()
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.REPLICATION_TRANS,
                                        authorInfoDict)
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnSrc], 'Processed')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailable], 'Deleted')

    # check moving transformation treatment for unused files
    self.fstAgent.setFileStatus.reset_mock()
    self.fstAgent.processTransformation(self.fakeTransID, self.sourceSE, self.targetSE, FST.MOVING_TRANS,
                                        authorInfoDict)
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailableOnSrc], 'Processed')
    self.fstAgent.setFileStatus.assert_any_call(self.fakeTransID, [fileNotAvailable], 'Deleted')


if __name__ == "__main__":
  SUITE = unittest.defaultTestLoader.loadTestsFromTestCase(TestFSTAgent)
  TESTRESULT = unittest.TextTestRunner(verbosity=3).run(SUITE)



"""Retrieve CalibrHCal from Calibration.txt file."""

from __future__ import print_function
from __future__ import absolute_import
import sys
from Helper_Functions import find_between


Calibration_File_And_Path = sys.argv[1]
Energy_To_Calibrate = sys.argv[2]
Initial_Calibration_Constant = sys.argv[3]
HCal_Barrel_Or_EndCap = sys.argv[4].lower()
Mean_Or_Calibration_Constant = sys.argv[5]

# Mean is the last value for mean of Gaussian fit written to the Calibration.txt file.

Mean = '1.0'

with open(Calibration_File_And_Path, 'r') as f:
    searchlines = f.readlines()
    for line in searchlines:
        if HCal_Barrel_Or_EndCap == 'barrel':
            if 'HCal Barrel Digi Mean' in line:
                Mean = float(find_between(line, ' : ', ' :'))
        elif HCal_Barrel_Or_EndCap == 'endcap':
            if 'HCal EndCap Digi Mean' in line:
                Mean = float(find_between(line, ' : ', ' :'))
        else:
            print('Please select Barrel or Endcap')

if Mean_Or_Calibration_Constant == 'Mean':
    print(Mean)

elif Mean_Or_Calibration_Constant == 'Calibration_Constant':
    print(float(Energy_To_Calibrate) * float(Initial_Calibration_Constant) / float(Mean))

    Calibration_Text = '_____________________________________________________________________________________' + '\n'
    Calibration_Text += 'HCal_Digi_Extract.py retrieving '

    if HCal_Barrel_Or_EndCap == 'barrel':
        Calibration_Text += 'CalibrHCalBarrel '
    elif HCal_Barrel_Or_EndCap == 'endcap':
        Calibration_Text += 'CalibrHCalEndCap '

    Calibration_Text += 'from digitisation program.' + '\n'
    Calibration_Text += 'KaonL Energy To Calibrate                          : ' + str(Energy_To_Calibrate) + ' /GeV\n'
    Calibration_Text += 'Initial Calibration Constant                       : ' + \
        str(Initial_Calibration_Constant) + '\n'
    Calibration_Text += 'CalibrHCal' + HCal_Barrel_Or_EndCap + \
        ' Mean                              : ' + str(Mean) + ' /GeV \n\n'
    Calibration_Text += 'CalibrHCal' + HCal_Barrel_Or_EndCap + '                                   : ' + \
        str(float(Energy_To_Calibrate) * float(Initial_Calibration_Constant) / float(Mean)) + '\n\n'

    with open(Calibration_File_And_Path, 'a') as myfile:
        myfile.write(Calibration_Text)

else:
    print('Please select Mean or Calibration_Constant to extract.')



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Akiyas script for replication."""
from __future__ import print_function
from __future__ import absolute_import
from DIRAC.Core.Base import Script
from six.moves import input
Script.parseCommandLine()

name_of_replication = 'Replicate_higgs_desy'
description = "Replicate higgs at 250GeV from KEK to DESY"
# below are meta data tags
evttype = 'higgs'  # or any other event type
detectormodel = None  # "ILD_o1_V2"#or any other
energy = "250"
machineparams = 'TDR_ws'
datatype = 'gen'  # this is either gen or REC

detectortype = None  # or 'ILD'

softwaretag = None  # or 'v01-16' or any other value
# fmask = ".+(/v01-16)+.+(DST)+."#this need to be adapted to select only the software and the file type wanted
# it matches anything that contains /v01-15-03 and DST

machine = 'ilc'  # This is alway ilc

source = 'KEK-SRM'  # source of the files
destination = 'DESY-SRM'  # destination of the files

# below should not change

from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient

from DIRAC.TransformationSystem.Client.Transformation import Transformation
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

from DIRAC.Core.Security.ProxyInfo import getProxyInfo

proxyinfo = getProxyInfo()
if not proxyinfo['OK']:
  print("Not allowed to create production, you need a ilc_prod proxy: use dirac-proxy-init -g ilc_prod")
  exit(1)
if 'group' in proxyinfo['Value']:
  group = proxyinfo['Value']['group']
  if not group == "ilc_prod":
    print("Not allowed to create production, you need a ilc_prod proxy: use dirac-proxy-init -g ilc_prod")
    exit(1)
else:
  print("Could not determine group, you do not have the right proxy: use dirac-proxy-init -g ilc_prod")
  exit(1)

fc = FileCatalogClient()

meta = {}
if evttype:
  meta['EvtType'] = evttype
if energy:
  meta['Energy'] = energy

meta['Machine'] = machine

if detectortype:
  meta['DetectorType'] = detectortype

meta['Datatype'] = datatype
if softwaretag:
  meta['SoftwareTag'] = softwaretag
if detectormodel:
  meta['DetectorModel'] = detectormodel
if machineparams:
  meta['MachineParams'] = machineparams

res = fc.findFilesByMetadata(meta)
if not res['OK']:
  print("Something went wrong when finding the files: %s" % res['Message'])
  exit(1)

lfns = res['Value']
if not len(lfns):
  print("No files found with this meta data query:%s" % meta)
  exit(1)

print("Example file: %s" % lfns[0])

answer = input('Proceed and submit replication? (Y/N): ')
if not answer.lower() in ('y', 'yes'):
  print("Canceled")
  exit(1)

trc = TransformationClient()
res = trc.getTransformationStats(name_of_replication)
if res['OK']:
  print("Replication with name %s already exists! Cannot proceed." % name_of_replication)
  exit(1)

Trans = Transformation()
Trans.setTransformationName(name_of_replication)
Trans.setDescription(description)
Trans.setLongDescription(description)
Trans.setType('Replication')
Trans.setPlugin('Broadcast')
# Trans.setFileMask(fmask)
Trans.setSourceSE(source)
Trans.setTargetSE(destination)
Trans.setInputDataQuery(meta)
res = Trans.addTransformation()
if not res['OK']:
  print("Failed to add Replication: %s" % res['Message'])
  exit(1)
Trans.setStatus("Active")
Trans.setAgentType("Automatic")



#
# Copyright (c) 2009-2022 CERN. All rights nots expressly granted are
# reserved.
#
# This file is part of iLCDirac
# (see ilcdirac.cern.ch, contact: ilcdirac-support@cern.ch).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# In applying this licence, CERN does not waive the privileges and
# immunities granted to it by virtue of its status as an
# Intergovernmental Organization or submit itself to any jurisdiction.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
#
"""Module to check the file contents.

:author: Ching Bon Lam
"""

from __future__ import absolute_import
import os

from string import Template

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Subprocess import shellCall
from ILCDIRAC.Core.Utilities.PrepareLibs import removeLibc
from ILCDIRAC.Workflow.Modules.ModuleBase import ModuleBase

__RCSID__ = '$Id$'
LOG = gLogger.getSubLogger(__name__)


class CheckCollections(ModuleBase):
  """Check the collections in a given slcio file."""

  def __init__(self):

    super(CheckCollections, self).__init__()

    self.STEP_NUMBER = ''
    self.args = ''
    #self.result      = S_ERROR()
    self.jobID = None
    self.applicationName = 'CheckCollections'
    # Step parameters

    self.InputFile = []
    self.collections = None

  def execute(self):
    """Run the thing."""
    # Get input variables

    result = self.resolveInputVariables()

    # Checks

    if not self.platform:
      result = S_ERROR('No ILC platform selected')

    if 'LCIO' not in os.environ:
      LOG.error("Environment variable LCIO was not defined, cannot do anything")
      result = S_ERROR("Environment variable LCIO was not defined, cannot do anything")

    if not result['OK']:
      LOG.error("Failed to resolve the input parameters:", self.result["Message"])
      return result

    removeLibc(os.path.join(os.environ["LCIO"], "lib"))

    # Setting up script

    LD_LIBRARY_PATH = os.path.join("$LCIO", "lib")
    if 'LD_LIBRARY_PATH' in os.environ:
      LD_LIBRARY_PATH += ":" + os.environ['LD_LIBRARY_PATH']

    PATH = "$LCIO/bin"
    if 'PATH' in os.environ:
      PATH += ":" + os.environ['PATH']

    scriptContent = Template('''
#!/bin/sh

#------------------------------------------------------------------------------#
# Dynamically generated script by CheckCollections module                      #
#------------------------------------------------------------------------------#

declare -x LD_LIBRARY_PATH=$LD_LIBRARY_PATH_
declare -x PATH=$PATH_

python <<PYTHONSCRIPT

import sys, subprocess

exitStatus = 0

for file in $files:

    cmdResult      = subprocess.Popen( ["lcio", "count", file], stdout=subprocess.PIPE ).communicate()[ 0 ]
    numberOfEvents = int( cmdResult.strip().split()[1] )

    cmdAnajobResult = subprocess.Popen( ["anajob", file], stdout=subprocess.PIPE ).communicate()[ 0 ]

    for collection in $collections:

        cmdResult           = subprocess.Popen( ["grep", "-c", collection], stdin=subprocess.PIPE, stdout=subprocess.PIPE ).communicate( cmdAnajobResult )[ 0 ]
        numberOfCollections = int( cmdResult.strip() )

        if numberOfEvents != numberOfCollections:

            print 'Inconsistency in %s: %i events vs %i collections (%s)' % ( file, numberOfEvents, numberOfCollections, collection )

            exitStatus = 1
            #sys.exit( exitStatus )

sys.exit( exitStatus )

PYTHONSCRIPT

declare -x appstatus=$$?
exit $$appstatus

''')

    scriptContent = scriptContent.substitute(LD_LIBRARY_PATH_=LD_LIBRARY_PATH,
                                              PATH_=PATH,
                                              files=self.InputFile,
                                              collections=self.collections
                                            )

    # Write script to file

    scriptPath = 'CheckCollections_%s_Run_%s' % (self.applicationVersion, self.STEP_NUMBER)

    if os.path.exists(scriptPath):
      os.remove(scriptPath)

    script = open(scriptPath, 'w')
    script.write(scriptContent)
    script.close()

    # Setup log file for application stdout

    if os.path.exists(self.applicationLog):
      os.remove(self.applicationLog)

    # Run code

    os.chmod(scriptPath, 0o755)

    command = '"./%s"' % (scriptPath)

    self.setApplicationStatus('CheckCollections %s step %s' % (self.applicationVersion, self.STEP_NUMBER))
    self.stdError = ''

    self.result = shellCall(0,
                             command,
                             callbackFunction=self.redirectLogOutput,
                             bufferLimit=20971520
                           )

    # Check results

    resultTuple = self.result['Value']
    status = resultTuple[0]

    LOG.info("Status after the application execution is %s" % str(status))

    return self.finalStatusReport(status)

  def applicationSpecificInputs(self):

    # Logfile

    if not self.applicationLog:
      self.applicationLog = 'CheckCollections_%s_Run_%s.log' % (self.applicationVersion, self.STEP_NUMBER)

    #
    self.InputFile = [os.path.basename(myfile) for myfile in self.InputFile]
    #

    if len(self.collections) == 0:
      return S_ERROR('No list of collections defined to check for.')

    #

    return S_OK('Parameters resolved')



"""Test Core.Utilities.FilenameEncoder.py."""

from __future__ import absolute_import
import unittest

from ILCDIRAC.Core.Utilities.FilenameEncoder import FilenameEncoder, decodeFilename, makeFilename


class FilenameEncoderTests(unittest.TestCase):
  """FilenameEncoderTests, test FilenameEncoder."""

  def setUp(self):
    """Make fake files for the test."""
    self.encoder = FilenameEncoder()
    self.maxDiff = None

  def test_makeSimNames(self):
    """test FilenameEncoder for stdhep file........................................................."""
    genfile = "E0500-TDR_ws.Pae_ell.Gwhizard-1.95.eB.pL.I37538.001.stdhep"
    fileitem = decodeFilename(genfile)

    refitem = {
        "E": "500-TDR_ws",
        "P": "ae_ell",
        "G": "whizard-1_95",
        "e": "B",
        "p": "L",
        "I": "37538",
        "n": "001",
        "F": "stdhep"}
    self.assertEqual(fileitem, refitem)

    fileitem.update({"s": "v01-14-01-p00", "m": "ILD_o1_v05", "d": "sim", "t": "7642", "j": "132"})
    diritem = {}
    diritem.update(fileitem)
    diritem.update({"B": "/ilc/prod/ilc/ild/test/temp1",         # Base directory
                     "T": "%8.8d" % (int(fileitem["t"])),        # Directory name for ProdID
                     "J": "%3.3d" % (int(fileitem["j"]) / 1000),  # Sub-directory for job
                     "C": "1f_3f"})                             # event class

    energy_machine = diritem["E"].split("-")
    diritem.update({"D": "SIM",                             # Meta value for data type
                     "w": energy_machine[0],                 # Energy
                     "o": energy_machine[1]})              # Mcahine parameters

    simfile = self.encoder.convert("sim", "file", fileitem)
    simdir = self.encoder.convert("sim", "dir", diritem)
    simmeta = self.encoder.convert("sim", "meta", diritem)

    reffile = "sv01-14-01-p00.mILD_o1_v05.E500-TDR_ws.I37538.Pae_ell.eB.pL.n001.d_sim_7642_132.slcio"
    refdir = "/ilc/prod/ilc/ild/test/temp1/sim/500-TDR_ws/1f_3f/ILD_o1_v05/v01-14-01-p00/00007642/000"

    refmeta = {'/ilc/prod/ilc/ild/test/temp1/sim': {'Datatype': 'SIM'},
                '/ilc/prod/ilc/ild/test/temp1/sim/500-TDR_ws': {'Energy': '500', 'MachineParams': 'TDR_ws'},
                '/ilc/prod/ilc/ild/test/temp1/sim/500-TDR_ws/1f_3f': {'EventClass': '1f_3f'},
                '/ilc/prod/ilc/ild/test/temp1/sim/500-TDR_ws/1f_3f/ILD_o1_v05': {'DetectorModel': 'ILD_o1_v05'},
                '/ilc/prod/ilc/ild/test/temp1/sim/500-TDR_ws/1f_3f/ILD_o1_v05/v01-14-01-p00': {'ILDConfig': 'v01-14-01-p00'},
                '/ilc/prod/ilc/ild/test/temp1/sim/500-TDR_ws/1f_3f/ILD_o1_v05/v01-14-01-p00/00007642': {'ProdID': '7642'},
                '/ilc/prod/ilc/ild/test/temp1/sim/500-TDR_ws/1f_3f/ILD_o1_v05/v01-14-01-p00/00007642/000': {'kJobNumber': '000'}}

    self.assertEqual(simfile, reffile)
    self.assertEqual(simdir, refdir)
    self.assertEqual(simmeta, refmeta)

  def test_makeRecDstNames(self):
    """Make Rec filenames from sim filename  ............................................................................"""
    simfile = "sv01-14-01-p00.mILD_o1_v05.E500-TDR_ws.I37538.Pae_ell.eB.pL.n001.d_sim_8642_832.slcio"
    recitem = decodeFilename(simfile)

    refitem = {
        "s": "v01-14-01-p00",
        "m": "ILD_o1_v05",
        "E": "500-TDR_ws",
        "I": "37538",
        "P": "ae_ell",
        "e": "B",
        "p": "L",
        "n": "001",
        "d": "sim",
        "t": "8642",
        "j": "832",
        "F": "slcio"}
    self.assertEqual(recitem, refitem)

    recitem.update({"r": "v01-16-p05_500", "d": "rec", "t": "8643", "j": "1854"})
    recfile = self.encoder.convert("rec", "file", recitem)

    diritem = {}
    diritem.update(recitem)
    diritem.update({"B": "/ilc/prod/ilc/ild/test/temp1",        # Base directory
                     "T": "%8.8d" % (int(recitem["t"])),        # Directory name for ProdID
                     "J": "%3.3d" % (int(recitem["j"]) / 1000),  # Sub-directory for job
                     "C": "1f_3f"})                            # event class

    energy_machine = diritem["E"].split("-")
    diritem.update({"D": "REC",                             # Meta value for data type
                     "w": energy_machine[0],                 # Energy
                     "o": energy_machine[1]})              # Mcahine parameters

    recdir = self.encoder.convert("rec", "dir", diritem)
    recmeta = self.encoder.convert("rec", "meta", diritem)

    reffile = "rv01-16-p05_500.sv01-14-01-p00.mILD_o1_v05.E500-TDR_ws.I37538.Pae_ell.eB.pL.n001.d_rec_8643_1854.slcio"
    refdir = "/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/1f_3f/ILD_o1_v05/v01-16-p05_500/00008643/001"

    refmeta = {'/ilc/prod/ilc/ild/test/temp1/rec': {'Datatype': 'REC'},
                '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws': {'Energy': '500', 'MachineParams': 'TDR_ws'},
                '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/1f_3f': {'EventClass': '1f_3f'},
                '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/1f_3f/ILD_o1_v05': {'DetectorModel': 'ILD_o1_v05'},
                '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/1f_3f/ILD_o1_v05/v01-16-p05_500': {'ILDConfig': 'v01-16-p05_500'},
                '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/1f_3f/ILD_o1_v05/v01-16-p05_500/00008643': {'ProdID': '8643'},
                '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/1f_3f/ILD_o1_v05/v01-16-p05_500/00008643/001': {'kJobNumber': '001'}}

    self.assertEqual(recfile, reffile)
    self.assertEqual(recdir, refdir)
    self.assertEqual(recmeta, refmeta)

    dstitem = {}
    dstitem.update(recitem)
    dstitem.update({"r": "v01-16-p05_500", "d": "dst", "t": "8645", "j": "711"})
    dstfile = self.encoder.convert("dst", "file", dstitem)
    self.assertEqual(
        dstfile,
        "rv01-16-p05_500.sv01-14-01-p00.mILD_o1_v05.E500-TDR_ws.I37538.Pae_ell.eB.pL.n001.d_dst_8645_711.slcio")

    dstdiritem = {}
    dstdiritem.update(diritem)
    dstdiritem.update({"d": dstitem["d"], "D": "DST", "t": dstitem["t"], "j": dstitem["j"],
                         "T": "%8.8d" % (int(dstitem["t"])),        # Directory name for ProdID
                         "J": "%3.3d" % (int(dstitem["j"]) / 1000)   # Sub-directory for job
                       })

    dstdir = self.encoder.convert("dst", "dir", dstdiritem)
    self.assertEqual(dstdir, "/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/1f_3f/ILD_o1_v05/v01-16-p05_500/00008645/000")

    dstmeta = self.encoder.convert("dst", "meta", dstdiritem)
    refmeta = {'/ilc/prod/ilc/ild/test/temp1/dst': {'Datatype': 'DST'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws': {'Energy': '500', 'MachineParams': 'TDR_ws'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/1f_3f': {'EventClass': '1f_3f'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/1f_3f/ILD_o1_v05': {'DetectorModel': 'ILD_o1_v05'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/1f_3f/ILD_o1_v05/v01-16-p05_500': {'ILDConfig': 'v01-16-p05_500'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/1f_3f/ILD_o1_v05/v01-16-p05_500/00008645': {'ProdID': '8645'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/1f_3f/ILD_o1_v05/v01-16-p05_500/00008645/000': {'kJobNumber': '000'}}
    self.assertEqual(dstmeta, refmeta)

  def test_makeRecDstNamesFromOldILDDiracSimFile(self):
    """Make Rec filenames from past dirac sim filename  ............................................................................"""
    simfile = "sv01-14-01-p00.mILD_o1_v05.E500-TDR_ws.I108687.P6f_xxxyyx.eL.pR_sim_6714_1.slcio"
    recitem = decodeFilename(simfile)

    refitem = {
        "s": "v01-14-01-p00",
        "m": "ILD_o1_v05",
        "E": "500-TDR_ws",
        "I": "108687",
        "P": "6f_xxxyyx",
        "e": "L",
        "p": "R",
        "d": "sim",
        "t": "6714",
        "j": "1",
        "F": "slcio"}
    self.assertEqual(recitem, refitem)
    self.assertNotIn("n", recitem)  # No stdhep filenumber information

    recitem.update({"r": "v01-16-p05_500", "d": "rec", "t": "8643", "j": "1854"})

    diritem = {}
    diritem.update(recitem)
    diritem.update({"B": "/ilc/prod/ilc/ild/test/temp1",        # Base directory
                     "T": "%8.8d" % (int(recitem["t"])),        # Directory name for ProdID
                     "J": "%3.3d" % (int(recitem["j"]) / 1000),  # Sub-directory for job
                     "C": "6f_xxWW"})                            # event class

    energy_machine = diritem["E"].split("-")
    diritem.update({"D": "REC",                             # Meta value for data type
                     "w": energy_machine[0],                 # Energy
                     "o": energy_machine[1]})              # Mcahine parameters

    fe = FilenameEncoder()
    # Apply special rule, because "n" infor. not available.
    fe.rules["rec"]["file"] = "r%r.s%s.m%m.E%E.I%I.P%P.e%e.p%p.d_%d_%t_%j.slcio"
    fe.rules["dst"]["file"] = "r%r.s%s.m%m.E%E.I%I.P%P.e%e.p%p.d_%d_%t_%j.slcio"

    recfile = fe.convert("rec", "file", recitem)
    recdir = fe.convert("rec", "dir", diritem)
    recmeta = fe.convert("rec", "meta", diritem)

    reffile = "rv01-16-p05_500.sv01-14-01-p00.mILD_o1_v05.E500-TDR_ws.I108687.P6f_xxxyyx.eL.pR.d_rec_8643_1854.slcio"
    refdir = "/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/6f_xxWW/ILD_o1_v05/v01-16-p05_500/00008643/001"
    refmeta = {'/ilc/prod/ilc/ild/test/temp1/rec': {'Datatype': 'REC'},
               '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws': {'Energy': '500', 'MachineParams': 'TDR_ws'},
               '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/6f_xxWW': {'EventClass': '6f_xxWW'},
               '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/6f_xxWW/ILD_o1_v05': {'DetectorModel': 'ILD_o1_v05'},
               '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/6f_xxWW/ILD_o1_v05/v01-16-p05_500': {'ILDConfig': 'v01-16-p05_500'},
               '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/6f_xxWW/ILD_o1_v05/v01-16-p05_500/00008643': {'ProdID': '8643'},
               '/ilc/prod/ilc/ild/test/temp1/rec/500-TDR_ws/6f_xxWW/ILD_o1_v05/v01-16-p05_500/00008643/001': {'kJobNumber': '001'}}

    self.assertEqual(recfile, reffile)
    self.assertEqual(recdir, refdir)
    self.assertEqual(recmeta, refmeta)

    dstitem = {}
    dstitem.update(recitem)
    dstitem.update({"r": "v01-16-p05_500", "d": "dst", "t": "8645", "j": "711"})
    dstfile = fe.convert("dst", "file", dstitem)

    dstdiritem = {}
    dstdiritem.update(diritem)
    dstdiritem.update({"d": dstitem["d"], "D": "DST", "t": dstitem["t"], "j": dstitem["j"],
                         "T": "%8.8d" % (int(dstitem["t"])),        # Directory name for ProdID
                         "J": "%3.3d" % (int(dstitem["j"]) / 1000)   # Sub-directory for job
                       })

    dstdir = fe.convert("dst", "dir", dstdiritem)
    dstmeta = fe.convert("dst", "meta", dstdiritem)

    reffile = "rv01-16-p05_500.sv01-14-01-p00.mILD_o1_v05.E500-TDR_ws.I108687.P6f_xxxyyx.eL.pR.d_dst_8645_711.slcio"
    refdir = "/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/6f_xxWW/ILD_o1_v05/v01-16-p05_500/00008645/000"
    refmeta = {'/ilc/prod/ilc/ild/test/temp1/dst': {'Datatype': 'DST'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws': {'Energy': '500', 'MachineParams': 'TDR_ws'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/6f_xxWW': {'EventClass': '6f_xxWW'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/6f_xxWW/ILD_o1_v05': {'DetectorModel': 'ILD_o1_v05'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/6f_xxWW/ILD_o1_v05/v01-16-p05_500': {'ILDConfig': 'v01-16-p05_500'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/6f_xxWW/ILD_o1_v05/v01-16-p05_500/00008645': {'ProdID': '8645'},
                '/ilc/prod/ilc/ild/test/temp1/dst/500-TDR_ws/6f_xxWW/ILD_o1_v05/v01-16-p05_500/00008645/000': {'kJobNumber': '000'}}

    self.assertEqual(dstfile, reffile)
    self.assertEqual(dstdir, refdir)
    self.assertEqual(dstmeta, refmeta)

    del fe

  def test_makeRecDstfromDBDSim(self):

    simfile = "sv01-14-01-p00.mILD_o1_v05.E1000-B1b_ws.I200006.P4f_ww_h.eL.pR-00262.slcio"

    fe = FilenameEncoder()
    # Apply special rule, because "n" infor. not available.
    fe.rules["rec"]["file"] = "r%r.s%s.m%m.E%E.I%I.P%P.e%e.p%p.d_%d_%t_%j.slcio"
    fe.rules["dst"]["file"] = "r%r.s%s.m%m.E%E.I%I.P%P.e%e.p%p.d_%d_%t_%j.slcio"

# Encode sim filename in order to build rec/dst files, directories
    recitem = decodeFilename(simfile)

    refitem = {"s": "v01-14-01-p00", "m": "ILD_o1_v05", "E": "1000-B1b_ws", "I": "200006", "P": "4f_ww_h",
                "e": "L", "p": "R", "j": "00262", "F": "slcio"}
    self.assertEqual(recitem, refitem)

    recitem["r"] = "v01-16-p03"  # ILDConfig version for Marlin
    recitem["d"] = "rec"  # data type
    recitem["t"] = "7654"  # Production ID
    recitem["j"] = "3232"  # Job number
    recdiritem = recitem
    recdiritem["B"] = "/ilc/prod/ilc/mc-dbd/ild"
    recdiritem["T"] = "%8.8d" % (int(recitem["t"]))
    recdiritem["J"] = "%3.3d" % (int(recitem["j"]) / 1000)
    recdiritem["C"] = "4f_WW_hadronic"
    recmetaitem = {}
    recmetaitem.update(recdiritem)
    recmetaitem["D"] = "REC"
    energy_machine = recdiritem["E"].split("-")             #
    recmetaitem["w"] = energy_machine[0]                 # Energy
    recmetaitem["o"] = energy_machine[1]                 # Mcahine parameters
#
    recfile = fe.convert("rec", "file", recitem)
    recdir = fe.convert("rec", "dir", recdiritem)
    recmeta = fe.convert("rec", "meta", recmetaitem)

    refrecfile = "rv01-16-p03.sv01-14-01-p00.mILD_o1_v05.E1000-B1b_ws.I200006.P4f_ww_h.eL.pR.d_rec_7654_3232.slcio"
    refrecdir = "/ilc/prod/ilc/mc-dbd/ild/rec/1000-B1b_ws/4f_WW_hadronic/ILD_o1_v05/v01-16-p03/00007654/003"

    refrecmeta = {'/ilc/prod/ilc/mc-dbd/ild/rec': {'Datatype': 'REC'},
                  '/ilc/prod/ilc/mc-dbd/ild/rec/1000-B1b_ws': {'Energy': '1000', 'MachineParams': 'B1b_ws'},
                  '/ilc/prod/ilc/mc-dbd/ild/rec/1000-B1b_ws/4f_WW_hadronic': {'EventClass': '4f_WW_hadronic'},
                  '/ilc/prod/ilc/mc-dbd/ild/rec/1000-B1b_ws/4f_WW_hadronic/ILD_o1_v05': {'DetectorModel': 'ILD_o1_v05'},
                  '/ilc/prod/ilc/mc-dbd/ild/rec/1000-B1b_ws/4f_WW_hadronic/ILD_o1_v05/v01-16-p03': {'ILDConfig': 'v01-16-p03'},
                  '/ilc/prod/ilc/mc-dbd/ild/rec/1000-B1b_ws/4f_WW_hadronic/ILD_o1_v05/v01-16-p03/00007654': {'ProdID': '7654'},
                  '/ilc/prod/ilc/mc-dbd/ild/rec/1000-B1b_ws/4f_WW_hadronic/ILD_o1_v05/v01-16-p03/00007654/003': {'kJobNumber': '003'}}

    self.assertEqual(recfile, refrecfile)
    self.assertEqual(recdir, refrecdir)
    self.assertEqual(recmeta, refrecmeta)

    del fe

  def test_getputRules(self):

    newrule = {"dstm": {"file": ""}}
    newrule["dstm"]["file"] = "r%r.s%s.m%m.E%E.I%I.P%P.e%e.p%p.n%n.slcio"
    newrule["dstm"]["dir"] = "%B/%d/%E/%C/%m/%r"
    newrule["dstm"]["meta"] = {"%B/%d": {"Datatype": "%D"},
                               "%B/%d/%E": {"Energy": "%w", "MachineParams": "%o"},
                               "%B/%d/%E/%C": {"EventClass": "%C"},
                               "%B/%d/%E/%C/%m": {"DetectorModel": "%m"},
                               "%B/%d/%E/%C/%m/%r": {"ILDConfig": "%r"}}
    self.encoder.defineRules(newrule, datatype="dstm")

    self.assertIn("dstm", self.encoder.rules)
    self.assertIn("file", self.encoder.rules["dstm"])
    self.assertIn("dir", self.encoder.rules["dstm"])
    self.assertIn("meta", self.encoder.rules["dstm"])

    dstmfilerule = self.encoder.getARule("dstm", category="file")
    self.assertEqual(dstmfilerule, newrule["dstm"]["file"])
    dstmrule = self.encoder.getARule("dstm")
    self.assertEqual(dstmrule, newrule["dstm"])

    self.encoder.defineRules("abcde", datatype="dstm", category="file")
    self.assertNotEqual(self.encoder.rules["dstm"], newrule["dstm"]["file"])

    self.assertIn("sim", self.encoder.rules)
    self.encoder.defineRules(newrule)
    self.assertNotIn("sim", self.encoder.rules)

  def test_makeFilename_double_percents(self):
    fileformat = "E%Energy%-%MachineParams%.P%P"
    filemeta = {"Energy": "250", "MachineParams": "TDR_ws", "P": "4f_WW_h"}
    convstr = makeFilename(fileformat, filemeta, preonly=False)
    self.assertEqual(convstr, "E250-TDR_ws.P%P")

    conv2nd = makeFilename(convstr, filemeta)
    self.assertEqual(conv2nd, "E250-TDR_ws.P4f_WW_h")


if __name__ == "__main__":
  SUITE = unittest.defaultTestLoader.loadTestsFromTestCase(FilenameEncoderTests)
  TESTRESULT = unittest.TextTestRunner(verbosity=2).run(SUITE)
